
/*
 * wimextract.c
 *
 * A program to create WIM archives.
 *
 * wimlib - Library for working with WIM files 
 *
 * Copyright (C) 2012 Eric Biggers
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#include "wimlib.h"
#include "config.h"
#include <stdarg.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>

#define MAX_IMAGES 32

static const struct option longopts[] = {
	{"image", required_argument, NULL, 'i'},
	{"boot-index", required_argument, NULL, 'b'},
	{"help", no_argument, NULL, 'h'},
	{"version", no_argument, NULL, 'V'},
	{NULL, 0, NULL, 0}
};

static void fatal_error(const char *msg, ...)
{
	fputs("WIMCREATE: ", stderr);
	va_list va;
	va_start(va, msg);
	vfprintf(stderr, msg, va);
	va_end(va);
	exit(1);
}

static void usage(int status)
{
	const char *s = 
	"Usage: wimcreate [-i DIR]... [FILE]\n"
	"Create a WIM archive.\n"
	"\n"
	"  -i, --image=DIR              Include the contents of DIR as an image in the\n"
	"                               WIM.  This option must be specified at least\n"
	"                               one time.  The images are included in the WIM\n"
	"                               in the order given.\n"
	"  -b, --boot-index=NUM         Set image NUM as the boot image.  By default, the\n"
	"                                 first image is used for this.\n"
	"  -c, --compression-type=TYPE  Compress files using TYPE compression. TYPE\n"
	"                                 currently defaults to 'none', and no other\n"
	"                                 TYPEs are supported.\n"
	"  -h, --help                   Display this information.\n"
	"  -V, --version                Display version information.\n"
	;
	fputs(s, stdout);
	exit(status);
}

static void version()
{
	char *s = 
	"wimcreate (%s) %s\n"
	"Report bugs to %s\n"
	;
	printf(s, PACKAGE_NAME, PACKAGE_STRING, PACKAGE_BUGREPORT);
	exit(0);
}

void *xmalloc(size_t len)
{
	void *p = malloc(len);
	if (p == NULL)
		fatal_error("Out of memory (tried to allocate %llu bytes)!\n", len);
	return p;
}

char *xstrdup(const char *s)
{
	size_t len = strlen(s);
	char *p = xmalloc(len + 1);
	memcpy(p, s, len);
	p[len] = '\0';
	return p;
}

int main(int argc, char **argv)
{
	int             c;
	WIMStruct      *w;
	char           *out_wim;
	unsigned        num_images = 0;
	unsigned        flags = WIM_FLAG_UNCOMPRESSED;
	char           *image_paths[MAX_IMAGES];
	unsigned        boot_idx = 1;

	while ((c = getopt_long(argc, argv, "i:b:c:hV", longopts, NULL)) != -1) {
		switch (c) {
		case 'i':
			if (num_images == MAX_IMAGES)
				fatal_error("Only up to %u images are supported!\n",
						MAX_IMAGES);
			else
				image_paths[num_images++] = xstrdup(optarg);
			break;
		case 'b':
			boot_idx = strtoul(optarg, NULL, 10);
			break;
		case 'c':
			if (strcmp(optarg, "none") != 0)
				fatal_error("Only `none' compression is "
						"currently supported!\n");
			break;
		case 'h':
			usage(0);
		case 'V':
			version();
		default:
			usage(1);
		}
	}
	argc -= optind;
	argv += optind;
	if (argc != 1)
		usage(1);
	if (num_images == 0)
		fatal_error("At least one image must be specified!\n");
	out_wim = argv[0];
	w = wimlib_create_new_wim(flags, boot_idx);
	for (int i = 0; i < num_images; i++) {
		if (!wimlib_add_image(w, image_paths[i])) {
			fatal_error("Failed to add the image `%s'\n", 
					image_paths[i]);
		}
	}
	if (!wimlib_write(w, out_wim, false))
		fatal_error("Failed to write wim file to `%s'\n", out_wim);
	return 0;
}

