/*
 * wimextract.c
 *
 * A program to extract files from WIM archives.
 *
 * wimlib - Library for working with WIM files 
 *
 * Copyright (C) 2012 Eric Biggers
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#include "wimlib.h"
#include "config.h"
#include <stdlib.h>
#include <getopt.h>
#include <string.h>

static const struct option longopts[] = {
	{"image", required_argument, NULL, 'i'},
	{"output", required_argument, NULL, 'o'},
	{"verbose", no_argument, NULL, 'v'},
	{"symbolic-links", no_argument, NULL, 's'},
	{"no-links", no_argument, NULL, 'n'},
	{"xml", required_argument, NULL, 'x'},
	{"help", no_argument, NULL, 'h'},
	{"version", no_argument, NULL, 'V'},
	{NULL, 0, NULL, 0}
};

static void usage(int status)
{
	const char *s = 
 	"Usage: wimextract [-o DIR] [-snvh]  [-x FILE] WIMFILE [FILE...]\n"
	"Extract file(s) from WIMFILE.\n"
	"\n"
	"  -i, --image=NUM      Select the image to extract files from.\n"
	"                         By default, all images are extracted.\n"
	"  -o, --output=DIR     Extract files into DIR.  DIR will be created\n"
	"                         if it does not exist, but the parent of DIR\n"
	"                         must exist.                                \n"
	"  -s, --symbolic-links Extract linked files in the WIM as symbolic links\n"
	"                         rather than hard links.\n"
	"  -n, --no-links       Extract linked files in the WIM as independent files,\n"
	"                         without links.\n" 
	"  -v, --verbose        Print the names of files as they are extracted.\n"
	"  -x, --xml=FILE       Extract the XML data to FILE.\n"
	"  -h, --help           Display this information.\n"
	"  -V, --version        Display version information.\n"
	"\n"
	"Examples:\n"
	"wimextract boot.wim\n"
	"  Extracts all files from \"boot.wim\" into the current directory.\n"
	"\n"
	"wimextract boot.wim \"Program Files\"\n"
	"  Extracts the directory \"Program Files\" from boot.wim into the\n"
	"  current directory.                                             \n"
	"\n"
	"wimextract -o winpe_base boot.wim\n"
	"  Extracts all files from \"boot.wim\" into the directory        \n"
	"  directory \"winpe_base\".                                      \n"
	;
	fputs(s, stdout);
	exit(status);
}

static void version()
{
	char *s = 
	"wimextract (%s) %s\n"
	"Report bugs to %s\n"
	;
	printf(s, PACKAGE_NAME, PACKAGE_STRING, PACKAGE_BUGREPORT);
	exit(0);
}


int main(int argc, char **argv)
{
	int c;

	WIMStruct *w = wimlib_create();
	const char *output_dir = ".";
	bool print_files = false;
	const char *xml_out_file = NULL;

	int link_type = WIM_LINK_TYPE_HARD;

	unsigned image = 0;

	while ((c = getopt_long(argc, argv, "i:o:vx:nshV", longopts, NULL)) != -1) {
		switch (c) {
		case 'i':
			image = strtoul(optarg, NULL, 10);
			break;
		case 'o':
			output_dir = strdup(optarg);
			if (output_dir == NULL) {
				fprintf(stderr, "Out of memory\n");
				return 1;
			}
			break;
		case 'n':
			link_type = WIM_LINK_TYPE_NONE;
			break;
		case 's':
			link_type = WIM_LINK_TYPE_SYMBOLIC;
			break;
		case 'v':
			print_files = true;
			break;
		case 'x':
			xml_out_file = strdup(optarg);
			if (xml_out_file == NULL) {
				fprintf(stderr, "Out of memory\n");
				return 1;
			}
			break;
		case 'h':
			usage(0);
		case 'V':
			version();
		default:
			usage(1);
		}
	}

	argv += optind;
	argc -= optind;

	if (argc < 1)
		usage(1);

	if (!wimlib_begin_read(w, argv[0])) {
		fprintf(stderr, "Failed to begin reading WIM file!\n");
		return 1;
	}

	if (print_files)
		wimlib_set_print_files(w);

	wimlib_set_link_type(w, link_type);

	if (!wimlib_set_output_dir(w, output_dir))
		return false;

	if (image != 0)
		if (!wimlib_select_image(w, image))
			return 1;

	if (argc == 1 && xml_out_file == NULL) {
		if (image != 0) {
			printf("WIMEXTRACT: Extracting files from image %u "
				"into directory `%s'\n", image, output_dir);
			if (!wimlib_extract_image(w)) {
				fprintf(stderr, "Failed to extract image %u!\n",
						image);
				return 1;
			}
		} else {
			printf("WIMEXTRACT: Extracting all files from `%s' "
				"into directory `%s'\n", argv[0], output_dir);
			if (!wimlib_extract_all_images(w)) {
				fprintf(stderr, "Failed to extract all files!\n");
				return 1;
			}
		}
	} else {
		for (int i = 1; i < argc; i++) {
			if (!wimlib_extract_tree(w, argv[i])) {
				fprintf(stderr, "Failed to extract the file `%s'!\n", argv[i]);
				return 1;
			}
		}
	}

	if (xml_out_file != NULL)
		if (!wimlib_extract_xml_data(w, xml_out_file))
			return 1;

	wimlib_end_read(w);
	wimlib_destroy(w);

	return 0;
}

