/*
 * wiminfo.c
 *
 * A program to print information from WIM files.
 *
 * wimlib - Library for working with WIM files 
 *
 * Copyright (C) 2012 Eric Biggers
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#include "wimlib.h"
#include "config.h"
#include <getopt.h>
#include <stdlib.h>

static const struct option longopts[] = {
	{"all", no_argument, NULL, 'a'},
	{"boot-metadata", no_argument, NULL, 'b'},
	{"files", no_argument, NULL, 'f'},
	{"image", required_argument, NULL, 'i'},
	{"lookup-table", no_argument, NULL, 'l'},
	{"header", no_argument, NULL, 'h'},
	{"xml", no_argument, NULL, 'x'},
	{"help", no_argument, NULL, 'H'},
	{"version", no_argument, NULL, 'V'},
	{NULL, 0, NULL, 0}
};

static void usage(int status)
{
	const char *s = 
	"Usage: wiminfo [OPTION]... [WIMFILE]\n"
	"Print information about WIMFILE.\n"
	"\n"
	"  -a, --all            Show all categories of information (default)\n"
	"  -h, --header         Show WIM header\n"
	"  -b, --boot-metadata  Show boot metadata for the selected image.\n"
	"  -f, --files          List all files in the selected image.\n"
	"  -i, --image=NUM      Select the image to show data for.  Used in\n"
	"                         -b and -f options.  Image 0 is the default.\n"
	"  -l, --lookup-table   Show lookup table\n"
	"  -x, --xml            Show XML data\n"
	"  -H, --help           Display this information.\n"
	"  -V, --version        Display version information.\n"
	;
	fputs(s, stdout);
	exit(status);
}

static void version()
{
	char *s = 
	"wiminfo (%s) %s\n"
	"Report bugs to %s\n"
	;
	printf(s, PACKAGE_NAME, PACKAGE_STRING, PACKAGE_BUGREPORT);
	exit(0);
}


int main(int argc, char **argv)
{
	int c;
	bool show_all           = false;
	bool show_header        = true;
	bool show_header_option = false;
	bool show_files         = false;
	bool show_boot_metadata = false;
	bool show_lookup_table  = false;
	bool show_xml           = false;

	WIMStruct *w = wimlib_create();

	unsigned image = 1;

	while ((c = getopt_long(argc, argv, "abfhi:lxHV", longopts, NULL)) != -1) {
		switch (c) {
		case 'a':
			show_all = true;
			show_header = false;
			break;
		case 'b':
			show_boot_metadata = true;
			show_header = false;
			break;
		case 'f':
			show_files = true;
			show_header = false;
			break;
		case 'h':
			show_header_option = true;
			break;
		case 'i':
			image = strtoul(optarg, NULL, 10);
			break;
		case 'l':
			show_lookup_table = true;
			show_header = false;
			break;
		case 'x':
			show_xml = true;
			show_header = false;
			break;
		case 'H':
			usage(0);
		case 'V':
			version();
		default:
			usage(1);
		}
	}

	argv += optind;
	argc -= optind;

	if (argc != 1)
		usage(1);

	if (show_all) {
		show_header        = true;
		show_boot_metadata = true;
		show_lookup_table  = true;
		show_xml           = true;
	}

	if (show_header_option)
		show_header = true;

	if (!wimlib_begin_read(w, argv[0]))
		return 1;

	if (!wimlib_select_image(w, image))
		return 1;


	if (show_header)
		if (!wimlib_print_wim_header(w))
			return 1;

	if (show_lookup_table)
		if (!wimlib_print_lookup_table(w))
			return 1;

	if (show_xml)
		if (!wimlib_print_xml_data(w))
			return 1;

	if (show_boot_metadata)
		if (!wimlib_print_boot_metadata(w))
			return 1;

	if (show_files)
		if (!wimlib_print_files(w))
			return 1;
	
	wimlib_end_read(w);
	wimlib_destroy(w);

	return 0;
}
