/* 
 * wimupdate.c
 *
 * This program is used to update WIM files.  It accepts commands reminiscent of
 * the UNIX shell, such as 'rm' and 'mv'.  It can be used interactively or
 * non-interactively.
 *
 * wimlib - Library for working with WIM files 
 *
 * Copyright (C) 2012 Eric Biggers
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 *
 */

#include "wimupdate.h"
#include "config.h"

#include <string.h>
#include <getopt.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>
#include <unistd.h>
#include <errno.h>
#include <sys/wait.h>


WIMStruct *w;
static bool interactive_mode;

static const struct option longopts[] = {
	{"actionfile", required_argument, NULL, 'a'},
	{"execute", required_argument, NULL, 'e'},
	{"help", no_argument, NULL, 'h'},
	{"version", no_argument, NULL, 'V'},
	{NULL, 0, NULL, 0}
};

static const struct option ls_longopts[] = {
	{"recursive", no_argument, NULL, 'r'},
	{NULL, 0, NULL, 0}
};

static const struct option mkdir_longopts[] = {
	{"parents", no_argument, NULL, 'p'},
	{NULL, 0, NULL, 0}
};

static const struct option mv_longopts[] = {
	{"verbose", no_argument, NULL, 'v'},
	{NULL, 0, NULL, 0}
};

static const struct option rm_longopts[] = {
	{"recursive", no_argument, NULL, 'r'},
	{"verbose", no_argument, NULL, 'v'},
	{NULL, 0, NULL, 0}
};

static void fatal_error(const char *msg, ...)
{
	fputs("WIMUPDATE: ", stderr);
	va_list va;
	va_start(va, msg);
	vfprintf(stderr, msg, va);
	va_end(va);
	exit(1);
}

void *xmalloc(size_t len)
{
	void *p = malloc(len);
	if (p == NULL)
		fatal_error("Out of memory (tried to allocate %llu bytes)!\n", len);
	return p;
}

static const char *path_basename(const char *path)
{
	long i = strlen(path) - 1;
	while (i >= 0 && path[i] == '/')
		i--;
	while (i >= 0 && path[i] != '/')
		i--;
	return path + i + 1;
}

char *xstrdup(const char *s)
{
	size_t len = strlen(s);
	char *p = xmalloc(len + 1);
	memcpy(p, s, len);
	p[len] = '\0';
	return p;
}

#define COLORS

#ifdef COLORS
#define GREEN "\e[32m"
#define BLUE "\e[33m"
#define RESET "\e[0m"
#else
#define GREEN ""
#define BLUE "" 
#define RESET ""
#endif


static void usage(int status)
{
	const char *s =
	"Usage: wimupdate [-a FILE]... [-e STRING] WIMFILE\n"
	"wimupdate uses wimlib to make updates to a WIM archive.  It can be\n"
	"used interactively or non-interactively.\n"
	"\n"
	"Noninteractive changes to be made are specified either in files given to\n"
	"the -a option, or in strings given to the -e argument.  If neither of these\n"
	"arguments are given, an interactive session is begun.  The command syntax for\n"
	"all these sources of commands is the same.\n"
	"\n"
	"The command syntax is similar to the UNIX shell.  Commands consist of a\n"
	"keyword, a number of arguments, and a command separator (semicolon or newline).\n"
	"Arguments may be quoted to cause a name such as \"Program Files\" to be\n"
	"interpreted as one argument.  WIMLIB maintains a working directory for\n"
	"the archive, and paths may be given relative to that working directory.\n"
	"The \"write\" command must be used for any changes to be saved.\n"
	"\n"
	"Options:\n"
	"  -a, --actionfile=FILE   Execute commands in FILE.\n"
	"  -e, --execute=STRING    Execute commands in STRING.\n"
	"  -h, --help              Display this information.\n"
	"  -V, --version           Display version information.\n"
	"\n"
	"The available commands are:\n"
	"\n"
#define DEFCMD(upper, lower, usage) \
	"  " #lower " " usage "\n"
#include "commands.def"
#undef DEFCMD
	"See the manual page for `wimupdate' for full documentation on the commands.\n"
	"\n"
	;
	fputs(s, stdout);
	exit(status);
}

static void version()
{
	char *s = 
	"wimupdate (%s) %s\n"
	"Report bugs to %s\n"
	;
	printf(s, PACKAGE_NAME, PACKAGE_STRING, PACKAGE_BUGREPORT);
	exit(0);
}


static struct list_node *new_list_node(char *str)
{
	struct list_node *new = xmalloc(sizeof(struct list_node));
	new->str = str;
	new->next = NULL;
	return new;
}

void list_add_string(struct list *list, char *str)
{
	struct list_node *new = new_list_node(str);
	if (list->first == NULL)
		list->first = new;
	else
		list->last->next = new;
	list->last = new;
}
void list_prepend_string(struct list *list, char *str)
{
	struct list_node *new = new_list_node(str);
	if (list->first == NULL)
		list->last = new;
	else
		new->next = list->first;
	list->first = new;
}

void list_add_string_from_file(struct list *list, const char *filename)
{
	off_t len;
	char *s = wimlib_file_get_contents(filename, &len);

	if (s == NULL)
		fatal_error("Error reading action file `%s': %s\n",
				filename, strerror(errno));
	list_add_string(list, s);
}

struct list *new_list()
{
	struct list *list = xmalloc(sizeof(struct list));
	list->first = NULL;
	list->last = NULL;
	return list;
}

static const char *command_names[] = {
#define DEFCMD(u, l, s) [ u ] = #l,
#include "commands.def"
#undef DEFCMD
};

static const char *usage_strings[] = {
#define DEFCMD(u, l, s) [ u ] = #l " " s,
#include "commands.def"
#undef DEFCMD
};


void print_available_commands()
{
	char *s = 
	"Available commands are:\n"
	"  ?\n"
	"  !\n"
#define DEFCMD(u, l, s) "  " #l " " s "\n"
#include "commands.def"
#undef DEFCMD
	;
	fputs(s, stdout);
}


static void command_usage(int command)
{
	printf("Usage: %s\n", usage_strings[command]);
}

static bool execute_abort_command(int argc, char **argv)
{
	abort();
	return true;
}

static bool execute_add_or_mv_command(int argc, char **argv, int command_num, 
		bool wimlib_func(WIMStruct *, const char *, const char *))
{
	optind = 1;
	bool verbose = false;
	int c;
	while ((c = getopt_long(argc, argv, "v", mv_longopts, NULL)) != -1) {
		switch (c) {
		case 'v':
			verbose = true;
			break;
		default:
			command_usage(command_num);
			return false;
			break;
		}
	}
	argc -= optind;
	argv += optind;
	if (argc > 2) {
		char *last = argv[argc - 1];
		if (!wimlib_is_directory(w, last)) {
			printf("%s: target `%s' is not a directory\n",
				command_names[command_num], last);
			return false;
		}
		for (int i = 0; i < argc - 1; i++) {
			if (!(*wimlib_func)(w, argv[i], last))
				return false;
			if (verbose) {
				printf("`%s' -> `%s/%s'\n", argv[i], last, 
							path_basename(argv[i]));
			}
		}
	} else if (argc == 2) {
		bool dest_is_dir = wimlib_is_directory(w, argv[1]);
		if (!(*wimlib_func)(w, argv[0], argv[1]))
			return false;
		if (verbose) {
			if (dest_is_dir)
				printf("`%s' -> `%s/%s'\n", argv[0], argv[1], 
							path_basename(argv[0]));
			else
				printf("`%s' -> `%s'\n", argv[0], argv[1]);
		}
	} else if (argc == 1) {
		printf("%s: missing destination file operand after `%s'\n",
			command_names[command_num], argv[0]);
		return false;
	} else {
		printf("%s: missing file operand\n", command_names[command_num]);
		return false;
	}
	return true;
}

static bool execute_add_command(int argc, char **argv)
{
	return execute_add_or_mv_command(argc, argv, ADD, wimlib_add);
}

static bool execute_cat_command(int argc, char **argv)
{
	argc--;
	argv++;
	bool ret = true;
	if (argc == 0) {
		printf("Usage: cat FILE...\n");
		ret = false;
	} else {
		for (int i = 0; i < argc; i++)
			 ret = ret && wimlib_cat(w, argv[i]);
	}
	return ret;
}

static bool execute_cd_command(int argc, char **argv)
{
	argc--;
	argv++;
	if (argc == 0)
		return wimlib_cd(w, "/");
	else
		return wimlib_cd(w, argv[0]);
}

static bool execute_exit_command(int argc, char **argv)
{
	if (wimlib_all_images_written(w))
		exit(0);
	else
		printf("Error: Modified WIM image(s) have not been saved!\n");
	return false;
}


static bool execute_help_command(int argc, char **argv)
{
	argc--;
	argv++;
	if (argc == 0) {
		print_available_commands();
		return true;
	} else {
		for (int i = 0; i < sizeof(command_names) / 
				sizeof(command_names[0]); i++)
		{
			if (strcmp(command_names[i], argv[0]) == 0) {
				command_usage(i);
				return true;
			}
		}
		return false;
			
	}
}

static bool execute_image_command(int argc, char **argv)
{
	argc--;
	argv++;
	if (argc == 0) {
		command_usage(IMAGE);
		return false;
	}
	return wimlib_select_image(w, strtoul(argv[0], NULL, 10));
}


static bool execute_ls_command(int argc, char **argv)
{
	optind = 1;
	int c;
	bool recursive = false;
	while ((c = getopt_long(argc, argv, "r", ls_longopts, NULL)) != -1) {
		switch (c) {
		case 'r':
			recursive = true;
			break;
		default:
			command_usage(LS);
			return false;
		}
	}
	argc -= optind;
	argv += optind;
	if (argc < 1)
		wimlib_ls(w, NULL, recursive);
	else
		for (int i = 0; i < argc; i++)
			wimlib_ls(w, argv[i], recursive);
	return true;
}

static bool execute_mkdir_command(int argc, char **argv)
{
	optind = 1;
	int c;
	bool parents = false;
	while ((c = getopt_long(argc, argv, "p", mkdir_longopts, NULL)) != -1) {
		switch (c) {
		case 'p':
			parents = true;
			break;
		default:
			command_usage(MKDIR);
			return false;
			break;
		}
	}
	argc -= optind;
	argv += optind;
	if (argc < 1) {
		command_usage(MKDIR);
		return false;
	}
	bool ret = true;
	for (int i = 0; i < argc; i++)
		ret = ret && wimlib_mkdir(w, argv[i], parents);
	return ret;
}

static bool execute_mv_command(int argc, char **argv)
{
	return execute_add_or_mv_command(argc, argv, MV, wimlib_mv);
}


static bool execute_rm_command(int argc, char **argv)
{
	optind = 1;
	bool recursive = false;
	bool verbose = false;
	int c;
	while ((c = getopt_long(argc, argv, "rv", rm_longopts, NULL)) != -1) {
		switch (c) {
		case 'r':
			recursive = true;
			break;
		case 'v':
			verbose = true;
			break;
		default:
			command_usage(RM);
			return false;
			break;
		}
	}
	argc -= optind;
	argv += optind;
	if (argc < 1) {
		command_usage(RM);
		return false;
	} else {
		bool ret = true;
		for (int i = 0; i < argc; i++)
			ret = ret && wimlib_rm(w, argv[i], recursive, verbose);
		return ret;
	}
}

static bool execute_shell_command(int argc, char **argv)
{
	argc--;
	argv++;
	if (argc == 0) {
		command_usage(SHELL);
		return false;
	}

	int total_len = 0;
	int lens[argc];
	for (int i = 0; i < argc; i++) {
		lens[i] = strlen(argv[i]);
		total_len += lens[i];
	}

	char cmd[total_len + 1];
	int j = 0;
	for (int i = 0; i < argc; i++) {
		memcpy(cmd + j, argv[i], lens[i]);
		j += lens[i];
	}
	cmd[j] = '\0';
		
	if (system(cmd) != 0) {
		printf("Error: %s\n", strerror(errno));
		return false;
	}
	return true;
}

static bool execute_show_command(int argc, char **argv)
{
	argc--;
	argv++;
	if (argc == 0) {
		command_usage(SHOW);
		return false;
	}
	if (strcmp(argv[0], "xml") == 0) {
		return wimlib_print_xml_data(w);
	} else if (strcmp(argv[0], "header") == 0) {
		return wimlib_print_wim_header(w);
	} else if (strcmp(argv[0], "num_images") == 0) { 
		printf("%u\n", wimlib_get_num_images(w));
		return true;
	} else if (strcmp(argv[0], "lookup_table") == 0) {
		return wimlib_print_lookup_table(w);
	} else if (strcmp(argv[0], "metadata") == 0) {
		return wimlib_print_boot_metadata(w);
	}
	
	command_usage(SHOW);
	return false;
}

static bool execute_stat_command(int argc, char **argv)
{
	argc--;
	argv++;
	if (argc != 1) {
		command_usage(STAT);
		return false;
	}
	return wimlib_stat(w, argv[0]);
}

static bool execute_write_command(int argc, char **argv)
{
	argc--;
	argv++;
	if (argc == 0) {
		command_usage(WRITE);
		return false;
	} else {
		bool current_image_only = false;
		if (argc >= 2) {
			current_image_only = true;
			if (!wimlib_select_image(w, strtoul(argv[0], NULL, 10)))
				return false;
			argv++;
		}
		if (!wimlib_write(w, argv[0], current_image_only)) {
			printf("Failed to write WIM file.\n");
			return false;
		}
	}
	return true;
}

#define DEFCMD(u, l, s) [ u ] = execute_##l##_command, 
static bool (*command_functions[])(int, char **) = {
	#include "commands.def"
};
#undef DEFCMD


/* Executes a command on the current WIM file.  The command word (RM, MKDIR,
 * etc.) is given in command->command_word.  The arguments are given as a linked
 * list of strings.  */
bool execute_command(struct list *command)
{
	char *argv[255];
	struct list_node *node = command->first;
	struct list_node *next;
	int argc = 1;
	while (node != NULL) {
		argv[argc] = node->str;
		if (++argc == 255)
			fatal_error("Too many arguments to command!\n");
		next = node->next;
		free(node);
		node = next;
	}
	return (*command_functions[command->command_word])(argc, argv);
}


/* Executes a linked list of actions on the current WIM file that contains
 * strings of commands read from the command line or from files. */
static bool execute_action_list(struct list *list)
{
	struct list_node *node = list->first;
	struct list_node *next;
	while (node != NULL) {
		if (!execute_action_string(node->str))
			return false;
		next = node->next;
		free(node->str);
		free(node);
		node = next;
	}
	return true;
}

static void read_commands_from_stdin()
{
	char buf[1000];
	printf("Welcome to WIMUPDATE!  Currently using the file `%s'\n", 
							wimlib_get_filename(w));
	print_available_commands();
	while (1) {
		printf(GREEN "%s (%u of %u): " BLUE "%s $ " RESET, 
				wimlib_get_filename(w),
				wimlib_get_current_image(w),
				wimlib_get_num_images(w),
				wimlib_get_working_directory_fullpath(w));

		if (fgets(buf, sizeof(buf), stdin) == NULL)
			break;
		if (!execute_action_string(buf))
			printf("Command did not complete successfully.\n");
	}
}

int yyerror(const char *s)
{
	printf("Unrecognized command\n");
	print_available_commands();
	return 0;
}

int main(int argc, char **argv)
{
	int c;

	struct list action_list = { NULL, NULL };

	w = wimlib_create();

	while ((c = getopt_long(argc, argv, "e:a:hV", longopts, NULL)) != -1) {
		switch (c) {
		case 'a':
			list_add_string_from_file(&action_list, optarg);
			break;
		case 'e':
			list_add_string(&action_list, xstrdup(optarg));
			break;
		case 'h':
			usage(0);
		case 'V':
			version();
		default:
			usage(1);
			break;
		}
	}

	argv += optind;
	argc -= optind;

	if (argc != 1)
		usage(1);

	const char *wim_in = argv[0];

	if (!wimlib_begin_read(w, wim_in)) {
		fatal_error("Failed to begin reading WIM file `%s'!\n", 
								wim_in);
	}

	if (action_list.first == NULL) {
		wimlib_select_image(w, 1);
		interactive_mode = true;
		read_commands_from_stdin();
	} else {
		interactive_mode = false;
		if (!execute_action_list(&action_list))
			return 1;
	}


	wimlib_end_read(w);
	wimlib_destroy(w);

	return 0;
}

