/*
 * create.c
 *
 * Support for creating WIM files.
 *
 * wimlib - Library for working with WIM files 
 *
 * Copyright (C) 2012 Eric Biggers
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#include "wimlib_internal.h"
#include <errno.h>

void create_header(WIMHeader *hdr, unsigned flags, unsigned boot_idx)
{
	hdr->magic[0] = 'M';
	hdr->magic[1] = 'S';
	hdr->magic[2] = 'W';
	hdr->magic[3] = 'I';
	hdr->magic[4] = 'M';
	hdr->magic[5] = '\0';
	hdr->magic[6] = '\0';
	hdr->magic[7] = '\0';
	hdr->hdr_size = WIM_HEADER_DISK_LEN;
	hdr->version  = 0x10d00;
	hdr->flags = 0;
	if (!(flags & WIM_FLAG_UNCOMPRESSED))
		hdr->flags &= WIM_HDR_FLAG_COMPRESSION;
	if (flags & WIM_FLAG_LZX)
		hdr->flags &= WIM_HDR_FLAG_COMPRESS_LZX;
	hdr->chunk_size = WIM_CHUNK_SIZE;
	randomize_byte_array(hdr->guid, sizeof(hdr->guid));
	hdr->part_number = 1;
	hdr->total_parts = 1;
	hdr->image_count = 0;
	hdr->boot_idx    = boot_idx;
	memset(&hdr->integrity, 0, sizeof(hdr->integrity));
}


static void init_security_data(WIMSecurityData *sd)
{
	sd->total_length = 8;
	sd->num_entries = 0;
}

/*
 * Creates a WIMStruct that contains no images.
 *
 * @flags:  A bitwise AND of the WIM_FLAG_* defined in wimlib.h.
 * @boot_idx:  The index of the image that will be the boot image.
 * @return:  A pointer to the new WIMStruct.
 */
WIMStruct *wimlib_create_new_wim(unsigned flags, unsigned boot_idx)
{
	DEBUG("Creating a new wim file (flags = %u, boot_idx = %u)\n",
			flags, boot_idx);
	/* Allocate the WIMStruct. */
	WIMStruct *w = wimlib_create();

	/* Create the header and lookup table. */
	create_header(&w->hdr, flags, boot_idx);
	lookup_table_init(&w->lookup_table, 9001);

	/* Make some dummy XML data, until real XML data is supported. */
	w->xml_data = xmalloc(3);
	w->xml_data[2] = '\0';
	w->hdr.xml_metadata.size = 2;
	w->hdr.xml_metadata.original_size = 2;
	w->hdr.xml_metadata.flags = 0;
	return w;
}

/* 
 * Adds an image to a WIM file.  The results are untested if this is done on a
 * WIMStruct that was not created by wimlib_create_new_wim().
 *
 * @w:  	The WIMStruct for the WIM file.
 * @path:  	The path to a directory in the outside filesystem; its contents will
 *  			be the contents of the image.
 * @return:	True on success, false on failure.
 */
bool wimlib_add_image(WIMStruct *w, const char *path)
{
	/* Insert an entry for the boot metadata into the lookup table. */
	DEBUG("Adding image with path `%s'\n", path);
	WIMLookupTableEntry *boot_metadata_lookup_entry = new_lookup_table_entry();
	randomize_byte_array(boot_metadata_lookup_entry->hash, WIM_HASH_SIZE);
	lookup_table_insert(&w->lookup_table, boot_metadata_lookup_entry);

	/* Reallocate the array of boot metadata. */
	w->hdr.image_count++;
	w->boot_metadata = xrealloc(w->boot_metadata, w->hdr.image_count * 
			sizeof(WIMBootMetadata));
	DEBUG("Reallocing boot metadata for image_count = %u\n",
			w->hdr.image_count);

	WIMBootMetadata *new_bmd = &w->boot_metadata[w->hdr.image_count - 1];
	WIMDentry *root_dentry = &new_bmd->root_dentry;

	/* Create the root dentry. */
	DEBUG("Creating root_dentry.\n");
	init_dentry(root_dentry, "");
	calculate_dentry_full_path(root_dentry);
	dentry_set_is_directory(root_dentry);
#ifdef ENABLE_DEBUG
	print_dentry(root_dentry, &w->lookup_table);
#endif

	/* Finish initializing the boot metadata. */
	init_security_data(&new_bmd->security_data);
	new_bmd->lookup_table_entry = boot_metadata_lookup_entry;
	new_bmd->modified = true;
	new_bmd->written = false;
	
	/* Change the current image to the new one. */
	if (!wimlib_select_image(w, w->hdr.image_count))
		return false;

	/* Add the dentry tree from disk. */
	struct stat root_stat;
	if (stat(path, &root_stat) != 0) {
		wimlib_error("Failed to stat `%s': %s\n", path, strerror(errno));
		free_dentry(root_dentry);
		return false;
	}
	if (!S_ISDIR(root_stat.st_mode)) {
		wimlib_error("`%s' is not a directory!\n", path);
		free_dentry(root_dentry);
		return false;
	}
	if (!build_dentry_tree(root_dentry, path, &root_stat, &w->lookup_table)) {
		wimlib_error("Failed to build dentry too for `%s'!\n", path);
		free_dentry_tree(root_dentry, &w->lookup_table, false, true, true);
		return false;
	}
	for_dentry_in_tree(&new_bmd->root_dentry, print_dentry, &w->lookup_table);
	return true;
}
