/*
 * dentry.c
 *
 * A dentry (directory entry) contains the metadata for a file.  In the WIM file
 * format, the dentries are stored in the "metadata resource" section, also
 * known as the "boot metadata", directly after the security data.
 *
 * wimlib - Library for working with WIM files 
 *
 * Copyright (C) 2010 Carl Thijssen
 * Copyright (C) 2012 Eric Biggers
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#include "wimlib_internal.h"

/* Arguments for do_free_dentry(). */
struct free_dentry_args {
	WIMDentry *root;
	WIMLookupTable *lookup_table;
	bool verbose;
	bool free_root;
	bool decrement_refcnt;
};

/* 
 * Calls a function on all directory entries in a directory tree.  It is called
 * on a parent before its children.
 *
 * @root: The root of the directory tree.
 * @visitor: The function to call. @args is passed as the second argument.  The
 * 	function must return true on success and false on failure, and the 
 * 	traversal is terminated if a failure result occurs. 
 * @args:  Second argument to @visitor.
 * @return:	True only if all invocations of @visitor returned true.
 */
INTERNAL bool for_dentry_in_tree(WIMDentry *root, 
		bool (*visitor)(WIMDentry*, void*), void *args)
{
	if (!(*visitor)(root, args))
		return false;
	WIMDentry *child = root->children;

	if (child == NULL)
		return true;
	do {
		if (!for_dentry_in_tree(child, visitor, args))
			return false;
		child = child->next;
	} while (child != root->children);
	return true;
}

/* 
 * Like for_dentry_in_tree(), but the visitor function is always called on a
 * dentry's children before on itself.
 */
INTERNAL bool for_dentry_in_tree_depth(WIMDentry *root, 
		bool (*visitor)(WIMDentry*, void*), void *args)
{

	WIMDentry *child = root->children;
	if (child != NULL) {
		do {
			if (!for_dentry_in_tree_depth(child, visitor, args))
				return false;
			child = child->next;
		} while (child != root->children);
	}
	if (!(*visitor)(root, args))
		return false;
	return true;
}

/* 
 * Calculate the full path of @dentry, based on its parent's full path and on
 * its ASCII file name. 
 */
INTERNAL void calculate_dentry_full_path(WIMDentry *dentry)
{
	int parent_len;
	int len;
	char *parent_fullpath;
	char *fullpath;
	if (dentry_is_root(dentry)) {
		dentry->fullpath_ascii = xmalloc(2);
		dentry->fullpath_ascii[0] = '/';
		dentry->fullpath_ascii[1] = '\0';
		dentry->fullpath_ascii_len = 1;
		return;
	}

	if (dentry_is_root(dentry->parent)) {
		parent_len = 0;
		parent_fullpath = "";
	} else {
		parent_len = dentry->parent->fullpath_ascii_len;
		parent_fullpath = dentry->parent->fullpath_ascii;
	}

	free(dentry->fullpath_ascii);
	len = parent_len + 1 + dentry->file_name_ascii_len;
	fullpath = xmalloc(len + 1);

	memcpy(fullpath, parent_fullpath, parent_len);
	fullpath[parent_len] = '/';
	memcpy(fullpath + parent_len + 1, dentry->file_name_ascii, 
				dentry->file_name_ascii_len);
	fullpath[len] = '\0';
	dentry->fullpath_ascii = fullpath;
	dentry->fullpath_ascii_len = len;
}

/* For use in for_dentry_in_tree(). */
INTERNAL bool do_calculate_dentry_full_path(WIMDentry *dentry, void *ignore)
{
	calculate_dentry_full_path(dentry);
	return true;
}


/*
 * Returns true if @dentry has the ASCII file name @name that has length
 * @name_len.
 */
INTERNAL bool dentry_has_name(const WIMDentry *dentry, 
			const char *name, int name_len)
{
	if (dentry->file_name_ascii_len != name_len)
		return false;
	return memcmp(dentry->file_name_ascii, name, name_len) == 0;
}

/* Returns the child of @dentry that has the file name @name.  
 * Returns NULL if no child has the name. */
INTERNAL WIMDentry *get_dentry_child_with_name(const WIMDentry *dentry, 
							const char *name)
{
	WIMDentry *child = dentry->children;
	if (child == NULL)
		return NULL;
	int len = strlen(name);
	do {
		if (dentry_has_name(child, name, len))
			return child;
		child = child->next;
	} while (child != dentry->children);
	return NULL;
}

/* Retrieves the dentry that has the ASCII @path relative to the dentry @cur_dir.  
 * . and ..  are handled properly. Returns NULL if no dentry having the path is
 * found. */
INTERNAL WIMDentry *get_dentry_relative_path(WIMDentry *cur_dir, const char *path)
{
	if (*path == '\0')
		return cur_dir;
	if (*path == '.') {
		if (*(path + 1) == '\0')
			return cur_dir;
		if (*(path + 1) == '/') {
			path += 2;
			while (*path == '/') 
				path++;
			return get_dentry_relative_path(cur_dir, path);
		} else if (*(path + 1) == '.') {
			if (*(path + 2) == '\0')
				return cur_dir->parent;
			if (*(path + 2) == '/') {
				if (cur_dir->parent == NULL)
					return NULL;
				path += 3;
				while (*path == '/')
					path++;
				return get_dentry_relative_path(cur_dir->parent, path);
			}
		}
	}
	WIMDentry *child = cur_dir->children;
	int base_len;
	const char *new_path = path_next_part(path, &base_len);
	if (child != NULL) {
		do {
			if (dentry_has_name(child, path, base_len))
				return get_dentry_relative_path(child, new_path);
			child = child->next;
		} while (child != cur_dir->children);
	}
	return NULL;
}

/* Retrieves the dentry that has the ASCII @path given as an absolute path.
 * @root is the root of the directory tree.  . and ..  are handled properly.
 * Returns NULL if no dentry having the path is found. */
static WIMDentry *get_dentry_absolute_path(WIMDentry *root, const char *path)
{
	if (*path == '\0')
		return root;
	while (*path == '/')
		path++;
	return get_dentry_relative_path(root, path);
}



/* Returns the dentry corresponding to the ASCII @path, which may be absolute or
 * relative to the working directory of the WIM struct. */
INTERNAL WIMDentry *get_dentry(WIMStruct *w, const char *path)
{
	if (is_absolute_path(path))
		return get_dentry_absolute_path(wim_root_dentry(w), path);
	else
		return get_dentry_relative_path(w->working_directory, path);
}

/* Returns the parent directory for the @path. */
INTERNAL WIMDentry *get_parent_dentry(WIMStruct *w, const char *path)
{
	DEBUG("Finding parent dentry for `%s'\n", path);
	int len = strlen(path);
	char buf[len + 1];
	memcpy(buf, path, len);

	to_parent_name(buf, len);

	if (is_absolute_path(path))
		return get_dentry_absolute_path(wim_root_dentry(w), buf);
	return get_dentry_relative_path(w->working_directory, buf);
}

/* Prints the full path of a dentry. */
INTERNAL bool print_dentry_full_path(WIMDentry *entry, void *ignore)
{
	puts(entry->fullpath_ascii);
	return true;
}

/* Prints a directory entry. */
INTERNAL bool print_dentry(WIMDentry *dentry, void *lookup_table)
{
	puts("[DIRECTORY ENTRY]");
	printf("Length            = %llu\n", dentry->length);
	printf("Attributes        = 0x%x\n", dentry->attributes);
	printf("Subdir offset     = %llu\n", dentry->subdir_offset);
	printf("Unused1           = %llu\n", dentry->unused1);
	printf("Unused2           = %llu\n", dentry->unused2);
	printf("Creation Time     = %llu\n", dentry->creation_time);
	printf("Last_Access Time  = %llu\n", dentry->last_access_time);
	printf("Last Write Time   = %llu\n", dentry->last_write_time);
	printf("Creation Time     = 0x%llx\n", dentry->creation_time);
	printf("Last_Access Time  = 0x%llx\n", dentry->last_access_time);
	printf("Last Write Time   = 0x%llx\n", dentry->last_write_time);
	printf("Hash              = "); 
	print_hash(dentry->hash); 
	putchar('\n');
	printf("Reparse Tag       = %u\n", dentry->reparse_tag);
	printf("Hard Link Group   = %llu\n", dentry->hard_link);
	printf("Number of Streams = %hu\n", dentry->streams);
	printf("Filename Length   = %hu\n", dentry->file_name_len);
	printf("Filename (ASCII) Length = %hu\n", dentry->file_name_ascii_len);
	printf("Short Name Length = %hu\n", dentry->short_name_len);
	printf("Short Name (ASCII) Length = %hu\n", dentry->short_name_ascii_len);
	printf("Filename          = \"");
	print_string(dentry->file_name, dentry->file_name_len);
	puts("\"");
	printf("Short Name        = \"");
	print_string(dentry->short_name, dentry->short_name_len);
	puts("\"");
	printf("Filename (ASCII)  = \"%s\"\n", dentry->file_name_ascii);
	if (dentry->short_name_ascii != NULL)
		printf("Short Name (ASCII)= \"%s\"\n", dentry->short_name_ascii);
	printf("Full Path (ASCII) = \"%s\"\n", dentry->fullpath_ascii);
	if (dentry->rest != NULL) {
		printf("Rest              = \"");
		print_string(dentry->rest, dentry->restsize);
		puts("\"");
	}
	printf("Rest Size         = %u\n", dentry->restsize);
	WIMLookupTableEntry *entry = lookup_file_resource(dentry, 
			(WIMLookupTable*)lookup_table);
	if (entry != NULL)
		print_lookup_table_entry(entry, NULL);
	return true;
}

/* 
 * Unlinks and frees a dentry tree.
 *
 * @root: 		The root of the tree.
 * @lookup_table:  	The lookup table for dentries.
 * @verbose:  		True if informational messages are to be printed.
 * @free_root:		True if the root is to be freed.
 * @decrement_refcnt:  	True if the dentries in the tree are to have their 
 * 			reference counts in the lookup table decremented.
 */
INTERNAL void free_dentry_tree(WIMDentry *root, WIMLookupTable *lookup_table, 
		bool verbose, bool free_root, bool decrement_refcnt)
{
	/*DEBUG("root = %p, lookup_table = %p, verbose = %d, free_root = %d,*/
			/*decrement_refcnt = %d\n", */
	struct free_dentry_args args;
	args.root                = root;
	args.lookup_table        = lookup_table;
	args.verbose             = verbose;
	args.free_root           = free_root;
	args.decrement_refcnt    = decrement_refcnt;
	for_dentry_in_tree_depth(root, do_free_dentry, &args);
}

INTERNAL void init_dentry(WIMDentry *dentry, const char *name)
{
	dentry->file_name = NULL;
	dentry->file_name_ascii = NULL;
	dentry->short_name_len = 0;
	dentry->short_name_ascii_len = 0;
	dentry->short_name = NULL;
	dentry->short_name_ascii = NULL;
	dentry->fullpath_ascii = NULL;
	dentry->rest = NULL;
	change_dentry_name(dentry, name);
	dentry->attributes = 0;
	dentry->security_id = 0;
	dentry->subdir_offset = 0;
	u64 t = get_timestamp();
	dentry->creation_time = t;
	dentry->last_access_time = t;
	dentry->last_write_time = t;
	memset(dentry->hash, 0, sizeof(dentry->hash));
	dentry->reparse_tag = 0;
	dentry->hard_link = 0;
	dentry->streams = 0;
	dentry->children = NULL;
	dentry->next = dentry;
	dentry->prev = dentry;
	dentry->parent = dentry;
}

/* 
 * Creates an unlinked directory entry.
 *
 * @name:    The base name of the new dentry.
 * @return:  A pointer to the new dentry.
 */
INTERNAL WIMDentry *create_dentry(const char *name)
{
	WIMDentry *dentry = xmalloc(sizeof(WIMDentry));
	init_dentry(dentry, name);
	return dentry;
}

/* Unlink a dentry from the directory tree. */
INTERNAL void unlink_dentry(WIMDentry *dentry)
{
	DEBUG("Unlinking `%s'\n", dentry->file_name_ascii);
	if (dentry_is_root(dentry))
		return;
	if (dentry_is_only_child(dentry)) {
		dentry->parent->children = NULL;
	} else {
		if (dentry_is_first_sibling(dentry))
			dentry->parent->children = dentry->next;
		dentry->next->prev = dentry->prev;
		dentry->prev->next = dentry->next;
	}
}

/* Frees memory allocated for a dentry (and inside the dentry). */
INTERNAL void free_dentry(WIMDentry *dentry)
{
	destroy_dentry(dentry);
	free(dentry);
}

/* 
 * This function is passed as an argument to for_dentry_in_tree_depth() in order
 * to free a directory tree.  __args is a pointer to a `struct free_dentry_args'.
 */
INTERNAL bool do_free_dentry(WIMDentry *dentry, void *__args)
{
	struct free_dentry_args *args = (struct free_dentry_args*)__args;
	if (args->verbose) {
		if (dentry_is_directory(dentry))
			printf("removed directory: `%s'\n", dentry->fullpath_ascii);
		else
			printf("removed `%s'\n", dentry->fullpath_ascii);
	}
	if (dentry == args->root) /* Only the root of the tree needs to be unlinked. */
		unlink_dentry(dentry);

	if (args->decrement_refcnt && !dentry_is_directory(dentry))
		decrement_ref_count(dentry, args->lookup_table);

	if (dentry != args->root || args->free_root)
		free_dentry(dentry);
	return true;
}





/* 
 * Links a dentry into the directory tree.
 *
 * @dentry: The dentry to link.
 * @parent: The dentry that will be the parent of @dentry.
 */
INTERNAL void link_dentry(WIMDentry *dentry, WIMDentry *parent)
{
	DEBUG("Linking `%s' with `%s'\n", dentry->file_name_ascii, 
				parent->fullpath_ascii);
	dentry->parent = parent;
	if (parent->children == NULL) {
		/* Only child; link to parent. */
		parent->children = dentry;
		dentry->next = dentry;
		dentry->prev = dentry;
	} else {
		/* Not an only child; link to siblings. */
		dentry->next = parent->children;
		dentry->prev = parent->children->prev;
		dentry->next->prev = dentry;
		dentry->prev->next = dentry;
	}
}

/* Frees memory allocated inside a dentry.  Does not free the dentry itself. */
INTERNAL void destroy_dentry(WIMDentry *dentry)
{
	free(dentry->file_name);
	free(dentry->file_name_ascii);
	free(dentry->short_name);
	free(dentry->short_name_ascii);
	free(dentry->fullpath_ascii);
	free(dentry->rest);
	for (int i = 0; i < dentry->streams; i++)
		free(dentry->stream_entries[i].stream_name);
	free(dentry->stream_entries);
}

/* Recalculates the length of @dentry based on its file name length and short
 * name length.  Frees the `rest' bytes if there are any. */
static void recalculate_dentry_size(WIMDentry *dentry)
{
	dentry->length = WIM_DENTRY_DISK_SIZE + dentry->file_name_len 
		+ WIM_FILENAME_PADDING + dentry->short_name_len;
	free(dentry->rest);
	dentry->rest = NULL;
	dentry->restsize = (8 - dentry->length % 8) % 8;
	dentry->length += dentry->restsize;
}

/* Changes the name of a dentry to @new_name.  Only changes the file_name and
 * file_name_ascii fields; does not change the short_name, short_name_ascii, or
 * fullpath_ascii fields.  Also recalculates its length. */
INTERNAL void change_dentry_name(WIMDentry *dentry, const char *new_name)
{
	free(dentry->file_name);
	free(dentry->file_name_ascii);
	int ascii_len = strlen(new_name);
	int wide_len;
	u8 *wide = ascii_to_wide(new_name, ascii_len, &wide_len);
	dentry->file_name = wide;
	dentry->file_name_len = wide_len;
	dentry->file_name_ascii = xmalloc(ascii_len + 1);
	dentry->file_name_ascii_len = ascii_len;
	memcpy(dentry->file_name_ascii, new_name, ascii_len + 1);
	recalculate_dentry_size(dentry);
}
