/*
 * extract.c
 *
 * Support for extracting WIM files.
 *
 * wimlib - Library for working with WIM files 
 *
 * Copyright (C) 2010 Carl Thijssen
 * Copyright (C) 2012 Eric Biggers
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#include "wimlib_internal.h"
#include "util.h"
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>



/* 
 * Extracts a regular file from the WIM archive. 
 *
 * @dentry:  		The directory entry for the file, which must be a
 *				regular file.
 * @output_path:   	The path to which the file is to be extracted.
 * @wimfile_p:  	The FILE* for the WIM file.
 * @wim_hdr_flags:  	The flags found in the header of the WIM file. 
 * @return: 		True on success, false on failure. 
 */
static bool extract_regular_file(const WIMDentry *dentry, 
		WIMLookupTable *lookup_table, const char *output_path, 
		FILE *wimfile_p, u32 wim_hdr_flags, int link_type)
{
	/* Lookup up the dentry in the lookup table. */
	WIMLookupTableEntry *entry = lookup_file_resource(dentry, lookup_table);

	/* If we already extracted the same file or a hard link copy of it, we
	 * may be able to simply create a link.  The exact action is specified
	 * by the current @link_type. */
	if (entry != NULL && entry->out_refcnt != 0 && 
					link_type != WIM_LINK_TYPE_NONE) {
		if (link_type == WIM_LINK_TYPE_HARD) {
			if (link(entry->file_on_disk, output_path) != 0) {
				wimlib_error("Failed to hard link `%s' to `%s': %s\n",
						output_path, entry->file_on_disk,
						strerror(errno));
				return false;
			}
		} else if (link_type == WIM_LINK_TYPE_SYMBOLIC) {
			int num_path_components = 
				get_num_path_components(dentry->fullpath_ascii);
			size_t file_on_disk_len = strlen(entry->file_on_disk);
			char buf[file_on_disk_len + 3 * num_path_components + 1];
			char *p = &buf[0];
			for (size_t i = 0; i < num_path_components; i++) {
				*p++ = '.';
				*p++ = '.';
				*p++ = '/';
			}
			strcpy(p, entry->file_on_disk);
			if (symlink(buf, output_path) != 0) {
				wimlib_error("Failed to symlink `%s' to `%s': %s\n",
						buf, entry->file_on_disk,
						strerror(errno));
				return false;
			}

		} else {
			assert(0);
		}
		return true;
	} 

	/* Otherwise, we must actually extract the file contents. */

	FILE *output_file = fopen(output_path, "wb");
	if (output_file == NULL) 
	{
		wimlib_error("Failed to open the file `%s' for writing: "
				"%s\n", output_path, strerror(errno));
		return false;
	}

	/* Extract empty file, with no lookup table entry... */
	if (entry == NULL) {
		fclose(output_file);
		return true;
	}
	const WIMResourceEntry *file_metadata = &entry->resource_entry;

	size_t size = file_metadata->original_size;
	u8 *file_contents = xmalloc(size);
	if (!read_file_resource(wimfile_p, file_metadata, wim_hdr_flags, 
				file_contents)) 
	{
		wimlib_error("Failed to read file resource `%s'!\n",
				dentry->fullpath_ascii);
		goto extract_regular_file_error;
	}

	if (fwrite(file_contents, 1, size, output_file) != size) {
		wimlib_error("Failed to write to output file `%s': %s!\n", 
				output_path, strerror(errno));
		goto extract_regular_file_error;
	}

	if (fclose(output_file) != 0) 
	{
		wimlib_error("Failed to close output file `%s': %s!\n",
				output_path, strerror(errno));
		goto extract_regular_file_error;
	}

	/* Mark the lookup table entry to indicate this file has been extracted.
	 * */
	entry->out_refcnt++;
	free(entry->file_on_disk);
	entry->file_on_disk = xstrdup(output_path);

	free(file_contents);
	return true;
extract_regular_file_error:
	free(file_contents);
	fclose(output_file);
	return false;
}

/* 
 * Extracts a directory from the WIM archive. 
 *
 * @dentry:		The directory entry for the directory.
 * @output_path:   	The path to which the directory is to be extracted to.
 * @return: 		True on success, false on failure. 
 */
static bool extract_directory(WIMDentry *dentry, const char *output_path)
{
	/* Compute the output path directory to the directory. */
	if (mkdir(output_path, S_IRWXU | S_IRGRP | S_IROTH | S_IXOTH) != 0) {
		switch (errno) {
		case EEXIST: /* Already existing directory is OK */
		case EACCES: /* We may have permissions to extract files inside
				 the directory, but not for the directory
				 itself. */
			return true;
		default:
			wimlib_error("Cannot create directory `%s': %s\n",
					output_path, strerror(errno));
			return false;
		}
	}
	return true;
}


/* 
 * Extracts a file or directory from the WIM archive.  For use in
 * for_dentry_in_tree().
 *
 * @dentry:	The dentry to extract.
 * @arg:	A pointer to the WIMStruct for the WIM file.
 */
static bool extract_regular_file_or_directory(WIMDentry *dentry, void *arg)
{
	WIMStruct *w = (WIMStruct*)arg;
	if (w->print_files)
		puts(dentry->fullpath_ascii);

	size_t len = strlen(w->output_dir);

	char path[len + dentry->fullpath_ascii_len + 1];

	memcpy(path, w->output_dir, len);
	memcpy(path + len, dentry->fullpath_ascii, dentry->fullpath_ascii_len);
	path[len + dentry->fullpath_ascii_len] = '\0';

	if (!dentry_is_directory(dentry)) {
		/* Regular file. */
		return extract_regular_file(dentry, &w->lookup_table,
				path, w->fp, w->hdr.flags, w->link_type);
	} else {
		/* Directory. */
		if (dentry_is_root(dentry)) /* Root doesn't need to be extracted. */
			return true;
		else
			return extract_directory(dentry, path);
	}
}


/* 
 * Extract the parent directories that are needed to extract @dir from the WIM.
 */
static bool extract_prerequisite_directories(WIMStruct *w, WIMDentry *dir)
{

	/* Compute the path that `dir' will be extracted to.  Then, go through
	 * each prefix of `buf' that is longer than the output root directory
	 * but shorter than the full path to `dir', and extract that directory.
	 * */
	int len1 = strlen(w->output_dir);
	int len2 = strlen(dir->fullpath_ascii);
	char buf[len1 + len2];
	memcpy(buf, w->output_dir, len1);
	memcpy(buf + len1, dir->fullpath_ascii, len2);
	to_parent_name(buf, len1 + len2);

	int name_idx = len1;
	while (1) {
		while (buf[name_idx] == '/')
			name_idx++;
		while (buf[name_idx] != '/' && buf[name_idx] != '\0')
			name_idx++;
		while (buf[name_idx] == '/')
			name_idx++;
		char save = buf[name_idx];
		buf[name_idx] = '\0';
		if (!extract_directory(NULL, buf))
			return false;
		buf[name_idx] = save;
		if (save == '\0')
			return true;
	}
}



/* Extracts the XML data of the WIM to the file outside the WIM named by
 * @out_file. */
bool wimlib_extract_xml_data(WIMStruct *w, const char *out_file)
{
	if (w->xml_data == NULL)
		if ((w->xml_data = read_xml_data(w->fp, &w->hdr.xml_metadata)) == NULL)
			return false;
	FILE *fp = fopen(out_file, "wb");
	if (fp == NULL) {
		wimlib_error("Failed to open the file `%s' for writing XML data!\n",
				out_file);
		return false;
	}
	if (fwrite(w->xml_data, 1, w->hdr.xml_metadata.size, fp) != 
			w->hdr.xml_metadata.size) 
	{
		wimlib_error("Failed to write XML data to the file `%s'!",
				out_file);
		return false;
	}
	return true;
}

static bool extraction_begun = false;

static inline bool begin_extract(WIMStruct *w)
{
	if (w->current_image == WIM_NO_IMAGE && !wimlib_select_image(w, 1)) {
		DEBUG("Failed to begin extraction!\n");
		return false;
	}
	lookup_table_zero_out_refcnts(&w->lookup_table);
	extraction_begun = true;
	return true;
}

/* Set the output directory for WIM extraction.  The directory is created using
 * mkdir() if it does not exist yet. */
bool wimlib_set_output_dir(WIMStruct *w, const char *dir)
{
	free(w->output_dir);
	w->output_dir = xstrdup(dir);

	DIR *d;
	if ((d = opendir(dir)) == NULL) {
		if (mkdir(dir, S_IRWXU | S_IRGRP | S_IROTH | S_IXOTH) != 0) {
			wimlib_error("Cannot create directory `%s': %s\n",
					dir, strerror(errno));
			return false;
		}
	} else {
		closedir(d);
	}
	return true;
}

/* Extracts the files from the WIM in the directory tree given by @path within
 * the currently selected image.  */
bool wimlib_extract_tree(WIMStruct *w, const char *path)
{
	if (!extraction_begun && !begin_extract(w))
		return false;
	WIMDentry *dentry = get_dentry(w, path);
	if (dentry == NULL) {
		wimlib_error("Could not find the dentry for `%s'\n", path);
		return false;
	}

	if (!extract_prerequisite_directories(w, dentry))
		return false;

	return for_dentry_in_tree(dentry, extract_regular_file_or_directory, w);
}

/* Extracts all the files from the currently selected image in the WIM. */
bool wimlib_extract_image(WIMStruct *w)
{
	DEBUG("Attempting to extract image %u...\n", wimlib_get_current_image(w));
	if (!extraction_begun && !begin_extract(w))
		return false;
	return for_dentry_in_tree(wim_root_dentry(w),
			extract_regular_file_or_directory, w);
}

void wimlib_set_link_type(WIMStruct *w, int link_type)
{
	w->link_type = link_type;
}

bool wimlib_extract_all_images(WIMStruct *w)
{
	DEBUG("Attempting to extract all images from `%s'\n",
			wimlib_get_filename(w));
	if (!extraction_begun && !begin_extract(w))
		return false;
	size_t output_path_len = strlen(w->output_dir);
	char buf[output_path_len + 10];
	memcpy(buf, w->output_dir, output_path_len);
	buf[output_path_len] = '/';
	for (unsigned i = 0; i < w->hdr.image_count; i++) {
		sprintf(buf + output_path_len + 1, "%u", i + 1);
		if (!wimlib_set_output_dir(w, buf))
			return false;
		if (!wimlib_select_image(w, i + 1))
			return false;
		if (!wimlib_extract_image(w))
			return false;
	}
	buf[output_path_len] = '\0';
	if (!wimlib_set_output_dir(w, buf))
		return false;
	return true;
}

