/*
 * lookup_table.c
 *
 * Lookup table that maps dentries to file resources.
 *
 * wimlib - Library for working with WIM files 
 *
 * Copyright (C) 2010 Carl Thijssen
 * Copyright (C) 2012 Eric Biggers
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#include "wimlib_internal.h"
#include <assert.h>

INTERNAL void lookup_table_init(WIMLookupTable *table, size_t capacity)
{
	table->num_entries = 0;
	table->capacity = capacity;
	table->array = xzalloc(capacity * sizeof(WIMLookupTableEntry*));
}

INTERNAL void lookup_table_destroy(WIMLookupTable *table)
{
	free(table->array);
}


INTERNAL WIMLookupTableEntry *new_lookup_table_entry()
{
	WIMLookupTableEntry *p = xmalloc(sizeof(WIMLookupTableEntry));
	p->file_on_disk = NULL;
	p->next = NULL;
	return p;
}

/*
 * Inserts an entry into the lookup table.
 *
 * @lookup_table:	A pointer to the lookup table.
 * @entry:		A pointer to the entry to insert.
 */
INTERNAL void lookup_table_insert(WIMLookupTable *table, 
					WIMLookupTableEntry *entry)
{
	WIMLookupTableEntry **array = table->array;
	size_t pos = *(u64*)entry->hash % table->capacity;

	DEBUG("Inserting %p into %p[%u]\n", entry, table->array, pos);
	if (array[pos] == NULL) {
		entry->next = NULL;
		array[pos] = entry;
	} else {
		entry->next = array[pos];
		array[pos] = entry;
	}
	table->num_entries++;
}

/* 
 * Looks up an entry in the lookup table.
 * 
 * @dentry:		The dentry for which to look up the entry.
 * @lookup_table:	A pointer to the lookup table.
 * @return:		A pointer to the WIMLookupTableEntry if found; otherwise
 * 				NULL.
 */
INTERNAL WIMLookupTableEntry *lookup_file_resource(const WIMDentry *dentry, 
						const WIMLookupTable *lookup_table)
{
	size_t pos = *(u64*)dentry->hash % lookup_table->capacity;
	DEBUG("Found file resource for `%s' in %p[%u]\n", dentry->file_name_ascii,
			lookup_table->array, pos);
	WIMLookupTableEntry *next = lookup_table->array[pos];
	while (next != NULL) {
		if (memcmp(dentry->hash, next->hash, WIM_HASH_SIZE) == 0)
			return next;
		next = next->next;
	}
	return NULL;
}



/* 
 * Calls a function on all the entries in the lookup table.  Stop early and
 * return false if any call to the function returns false.
 *
 * @table:	A pointer to the lookup table.
 * @visitor:	The function to call.
 * @arg:	An argument to pass to the function.
 * @return:	True if the function returns true every time it is called;
 * 			otherwise, false.
 */
INTERNAL bool for_lookup_table_entry(WIMLookupTable *table, 
		bool (*visitor)(WIMLookupTableEntry *, void *), void *arg)
{
	for (size_t i = 0; i < table->capacity; i++) {
		WIMLookupTableEntry *entry = table->array[i];
		while (entry != NULL) {
			if (!(*visitor)(entry, arg))
				return false;
			entry = entry->next;
		}
	}
	return true;
}

INTERNAL void free_lookup_table_entry(WIMLookupTableEntry *entry)
{
	free (entry->file_on_disk);
	free (entry);
}

/* Decrement the reference count for @dentry in the lookup table.  The lookup
 * table entry is freed if there are no references remaining. */
INTERNAL void decrement_ref_count(const WIMDentry *dentry, 
			WIMLookupTable* lookup_table)
{
	size_t pos = *(u64*)dentry->hash % lookup_table->capacity;
	WIMLookupTableEntry *entry = lookup_table->array[pos];
	WIMLookupTableEntry *prev = NULL;
	while (entry != NULL) {
		if (memcmp(dentry->hash, entry->hash, WIM_HASH_SIZE) == 0) {
			assert(entry->refcnt != 0);
			DEBUG("'%s': %u => %u\n", dentry->fullpath_ascii, 
					entry->refcnt, entry->refcnt - 1);
			if (--entry->refcnt == 0) {
				if (prev == NULL)
					lookup_table->array[pos] = NULL;
				else
					prev->next = entry->next;
				free_lookup_table_entry(entry);
			}
		}
			
		prev = entry;
		entry = entry->next;
	}
}

/*
 * Prints an entry from the lookup table.
 *
 * @entry:	The lookup table entry.
 * @ignore:	Ignored.
 * @return:	Always true.
 */
INTERNAL bool print_lookup_table_entry(WIMLookupTableEntry *entry, void *ignore)
{
	printf("Offset            = %llu bytes\n", 
			entry->resource_entry.offset);
	printf("Size              = %llu bytes\n", 
			entry->resource_entry.size);
	printf("Original size     = %llu bytes\n", 
			entry->resource_entry.original_size);
	printf("Part Number       = %hu\n", entry->part_number);
	printf("Reference Count   = %u\n", entry->refcnt);
	printf("Hash              = ");
	print_hash(entry->hash);
	putchar('\n');
	printf("Flags             = ");
	u8 flags = entry->resource_entry.flags;
	if (flags & WIM_RESHDR_FLAG_COMPRESSED)
		fputs("WIM_RESHDR_FLAG_COMPRESSED, ", stdout);
	if (flags & WIM_RESHDR_FLAG_FREE)
		fputs("WIM_RESHDR_FLAG_FREE, ", stdout);
	if (flags & WIM_RESHDR_FLAG_METADATA)
		fputs("WIM_RESHDR_FLAG_METADATA, ", stdout);
	if (flags & WIM_RESHDR_FLAG_SPANNED)
		fputs("WIM_RESHDR_FLAG_SPANNED, ", stdout);
	putchar('\n');
	if (entry->file_on_disk != NULL)
		printf("File on Disk      = \"%s\"\n", entry->file_on_disk);
	putchar('\n');
	return true;
}

/*
 * Prints the lookup table from the WIM file associated with @w.  Returns true on
 * success, false on failure.
 */
bool wimlib_print_lookup_table(WIMStruct *w)
{
	for_lookup_table_entry(&w->lookup_table, print_lookup_table_entry, NULL);
	return true;
}


