/*
 * modify.c
 *
 * Support for modifying WIM files.
 *
 * wimlib - Library for working with WIM files 
 *
 * Copyright (C) 2010 Carl Thijssen
 * Copyright (C) 2012 Eric Biggers
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#include "wimlib_internal.h"
#include "util.h"
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <sys/stat.h>
#include <dirent.h>



static void transfer_file_attributes(struct stat *stat_buf, WIMDentry *dentry)
{
	dentry->attributes = 0;
	if (S_ISDIR(stat_buf->st_mode))
		dentry_set_is_directory(dentry);
	else if (S_ISREG(stat_buf->st_mode))
		dentry_set_is_regular_file(dentry);
	else
		wimlib_error("Unknown st_mode\n");
}



/* 
 * Recursively builds a dentry tree from a directory tree on disk, outside the
 * WIM file.
 *
 * @root:  A dentry that has already been created for the root of the dentry
 * 	tree.
 * @source_path:  The path to the root of the tree on disk. 
 * @root_stat:   A pointer to a `struct stat' that contains the metadata for the
 * 	root of the tree on disk. 
 * @return:	True on success, false on failure.  It is a failure if any of
 *		the files cannot be `stat'ed, or if any of the needed
 *		directories cannot be opened or read.  Failure to add the files
 *		to the WIM may still occur later when trying to actually read 
 *		the regular files in the tree into the WIM as file resources.
 */
INTERNAL bool build_dentry_tree(WIMDentry *root, const char *source_path, 
			struct stat *root_stat, WIMLookupTable* lookup_table)
{
	transfer_file_attributes(root_stat, root);

	if (dentry_is_directory(root)) {

		/* Open the directory on disk */
		DIR *dir = opendir(source_path);
		if (dir == NULL) {
			wimlib_error("Failed to open the directory `%s': %s\n",
					source_path, strerror(errno));
			return false;
		}

		/* Buffer for names of files in directory. */
		int len = strlen(source_path);
		char name[len + 1 + FILENAME_MAX + 1];
		memcpy(name, source_path, len);
		name[len] = '/';
		struct dirent *p;
		errno = 0;

		/* Create a dentry for each entry in the directory on disk, and recurse
		 * to any subdirectories. */
		while ((p = readdir(dir)) != NULL) {
			if (p->d_name[0] == '.' && (p->d_name[1] == '\0'
				|| (p->d_name[1] == '.' && p->d_name[2] == '\0')))
					continue;
			strcpy(name + len + 1, p->d_name);
			struct stat child_stat;
			if (stat(name, &child_stat) != 0) {
				wimlib_error("cannot stat `%s': %s\n", name,
						strerror(errno));
				closedir(dir);
				return false;
			}
			WIMDentry *child = create_dentry(p->d_name);
			if (!build_dentry_tree(child, name, &child_stat, lookup_table)) {
				closedir(dir);
				return false;
			}
			link_dentry(child, root);
		}
		closedir(dir);
		if (errno != 0) {
			wimlib_error("Failed to read the directory `%s': %s\n",
					name, errno);
			return false;
		}
	} else {
		/* For each non-directory, we must check to see if the file is
		 * in the lookup table already; if it is, we increment its
		 * refcnt; otherwise, we create a new lookup table entry and
		 * insert it. */
		if (!sha1sum(source_path, root->hash)) {
			DEBUG("Failed to calculate sha1sum for file `%s'\n", 
						source_path);
			return false;
		}
		#ifdef ENABLE_DEBUG
		DEBUG("Calculated sha1sum for file `%s': ", source_path);
		print_byte_field(root->hash, WIM_HASH_SIZE);
		putchar('\n');
		#endif

		DEBUG("Looking up entry for `%s'\n", root->file_name_ascii);
		WIMLookupTableEntry *entry = lookup_file_resource(root, 
								lookup_table);
		if (entry == NULL) {
			DEBUG("Adding lookup table entry for `%s'\n", source_path);
			entry               = new_lookup_table_entry();
			entry->file_on_disk = xstrdup(source_path);
			entry->refcnt       = 1;
			entry->part_number  = 1;
			memcpy(entry->hash, root->hash, WIM_HASH_SIZE);
			lookup_table_insert(lookup_table, entry);
		} else {
			DEBUG("Found existing lookup table entry for `%s'\n", 
								source_path);
			entry->refcnt++;
		}
	}
	return true;
}

/* Moves a dentry into a directory.
 *
 * @dentry:	The dentry to move, which must currently be linked in some other
 * 		place, or else come from create_dentry() as linked to itself.
 * @dir:	The dentry for the directory.  It must be a directory.
 * @new_name:	If non-NULL, a pointer to an ASCII string that gives a new name
 * 		that will be given to @dentry.
 * @return	True on success, false on failure.  Possible failures include
 * 		attempting to move a directory into itself, or trying to
 * 		overwrite a subdirectory.
 */
static bool move_dentry_to_directory(WIMDentry *src_dentry, WIMDentry *dst_dir, 
				const char *new_name, WIMLookupTable *lookup_table)
{
	if (src_dentry == dst_dir) {
		wimlib_error("Cannot move `%s' to a subdirectory "
				"of itself!\n", src_dentry->file_name_ascii);
		return false;
	}

	const char *name = (new_name != NULL) ? new_name : src_dentry->file_name_ascii;
	DEBUG("Moving `%s' to `%s/%s'\n", src_dentry->file_name_ascii, 
			dst_dir->fullpath_ascii, name);

	WIMDentry *existing = get_dentry_child_with_name(dst_dir, name);
	if (existing != NULL && dentry_is_directory(existing)) {
		wimlib_error("Cannot overwrite directory `%s' with %s `%s'!\n",
				existing->fullpath_ascii, 
				dentry_is_directory(src_dentry) ? "directory" : "file",
				new_name);
		return false;
	}
	if (existing != NULL) {
		DEBUG("Unlinking existing file `%s'\n", existing->fullpath_ascii);
		unlink_dentry(existing);
		decrement_ref_count(existing, lookup_table);
		free_dentry(existing);
	}
	unlink_dentry(src_dentry);
	link_dentry(src_dentry, dst_dir);
	if (new_name != NULL)
		change_dentry_name(src_dentry, new_name);
	return true;
}

/* 
 * Inserts a dentry into the directory tree.  Missing directories are
 * automatically created.
 *
 * @tree:    	The root of the tree.
 * @dentry:  	The dentry to insert.
 * @path:	The path to insert the dentry into, relative to @tree.
 */
static bool insert_dentry(WIMDentry *tree, WIMDentry *dentry, 
						const char *path)
{
	DEBUG("Inserting %s into %s\n", path, tree->fullpath_ascii);
	int base_len;
	const char *new_path = path_next_part(path, &base_len);

	if (path[0] == '.') {
		if (base_len == 1)
			return insert_dentry(tree, dentry, new_path);
		else if (base_len == 2 && path[1] == '.')
			return insert_dentry(tree->parent, dentry, new_path);
	}

	if (*new_path == '\0') {
		/* In this case, @tree is a leaf entry, so link this dentry as a
		 * leaf. */
		if (!dentry_is_directory(tree)) {
			wimlib_error("Cannot overwrite non-directory `%s' with "
					"a directory!\n", tree->fullpath_ascii);
			return false;
		}
		link_dentry(dentry, tree);
		calculate_dentry_full_path(dentry);
		return true;
	} else {
		/* Linking as a nonleaf.  Look to see if there is a directory we
		 * need to go into. */
		WIMDentry *child = tree->children;
		if (child != NULL) {
			do {
				/* Yes, there is a directory we need to go into. */
				if (dentry_has_name(child, path, base_len))
					return insert_dentry(child, dentry, new_path);

				child = child->next;
			} while (child != tree->children);
		}

		/* No child has the name corresponding to the first part of the
		 * path, so we must create a new directory. */
		char s[base_len + 1];
		memcpy(s, path, base_len);
		s[base_len] = '\0';


		child = create_dentry(s);
		dentry_set_is_directory(child);

		link_dentry(child, tree);
		calculate_dentry_full_path(child);

		return insert_dentry(child, dentry, new_path);
	}
}

/* 
 * Moves a dentry to a path.  The dentry is linked in the new location, and the
 * full paths for it and its subtree are recalculated.
 *
 * @src_dentry:	The dentry to move, which must currently be linked in 
 * 			some other place, or else come from create_dentry() as 
 * 			linked to itself. It may be either a regular file or
 * 			directory.
 * @dest:	The destination path.
 * @w:		Pointer to the WIMStruct for the WIM.
 * @return:	True if the dentry was successfully moved.  This function
 * 		can fail for a number of reasons, including moving to a
 * 		destination with no existing parent directory, moving a dentry
 * 		to itself, overwriting a directory with a regular file, etc.
 */
static bool move_dentry_to_path(WIMDentry *src_dentry, const char *dest, 
							WIMStruct *w)
{
	WIMDentry *dst = get_dentry(w, dest);
	bool ret;
	const char *src_name = src_dentry->file_name_ascii;
	DEBUG("Moving `%s' to `%s'\n", src_name, dest);
	if (dst == NULL) {
		/* Destination path does not exist.  Get the parent directory.
		 * */
		WIMDentry *parent = get_parent_dentry(w, dest);
		DEBUG("Destination path `%s' does not exist.\n", dest);
		if (parent != NULL && dentry_is_directory(parent)) {
			/* If there is a parent, move the src dentry to that
			 * directory. */
			ret = move_dentry_to_directory(src_dentry, parent, 
					path_basename(dest), &w->lookup_table);
		} else {
			/* The parent of the destination is non existent, which
			 * is an error. */
			wimlib_error("Destination `%s' does not exist!\n", dest);
			ret = false;
		}
	} else {
		/* The destination path exists. */
		if (dentry_is_directory(dst)) {
			/* Move into existing directory. */
			ret = move_dentry_to_directory(src_dentry, dst, NULL, 
					&w->lookup_table);
		} else {
			/* Overwrite regular file. */
			if (src_dentry == dst) {
				wimlib_error("`%s' and `%s' are the same file\n", 
					src_name, dest);
				return false;
			}
			ret = move_dentry_to_directory(src_dentry, dst->parent, 
					path_basename(dest), &w->lookup_table);
		}
	}
	if (ret) {
		DEBUG("Calculating full paths for dentry tree rooted at `%s'\n",
				src_dentry->file_name_ascii);
		ret = for_dentry_in_tree(src_dentry, do_calculate_dentry_full_path, NULL);
	}

	return ret;
}

/* 
 * Creates and links a dentry for a file located outside the WIM.  
 *
 * @w:  The WIMStruct for the WIM file.
 * @source:  The path to the file outside the WIM.  It can be a regular file or
 * 	a directory.  If it is a directory, its contents are added recursively.
 * @dest:  The destination path inside the WIM.
 * @return:  True if the dentry(s) were successfully added.  This does not
 * 	necessarily mean that the files will be successfully added when
 * 	wimlib_write() is called because this step only adds the metadata.
 */
bool wimlib_add(WIMStruct *w, const char *source, const char *dest)
{
	DEBUG("Adding `%s' to path `%s'\n", source, dest);
	if (w->current_image == WIM_NO_IMAGE && !wimlib_select_image(w, 1))
		return false;
	/* Build the directory tree to be added. */
	struct stat buf;
	if (stat(source, &buf) != 0) {
		wimlib_error("Failed to stat `%s': %s\n", source, 
						strerror(errno));
		return false;
	}

	WIMDentry *root = create_dentry(path_basename(source));

	if (!build_dentry_tree(root, source, &buf, &w->lookup_table)) {
		free_dentry_tree(root, &w->lookup_table, false, true, true);
		return false;
	}

	if (!move_dentry_to_path(root, dest, w)) {
		free_dentry_tree(root, &w->lookup_table, false, true, true);
		return false;
	}
	wim_set_current_image_modified(w);
	return true;
}

/* 
 * Moves a file or directory within the WIM. 
 *
 * @w:		Pointer to the WIMStruct for the WIM.
 * @source:	Path to the source.
 * @dest:	Path to the destination.
 * @return:	True if the file was successfully moved.
 */
bool wimlib_mv(WIMStruct *w, const char *source, const char *dest)
{
	if (w->current_image == WIM_NO_IMAGE && !wimlib_select_image(w, 1))
		return false;
	WIMDentry *src = get_dentry(w, source);
	if (src == NULL) {
		wimlib_error("cannot stat `%s': No such file or directory\n",
					source);
		return false;
	}

	wim_set_current_image_modified(w);
	return move_dentry_to_path(src, dest, w);
}

/*
 * Makes a directory inside the WIM.
 *
 * @w:	Pointer to the WIMStruct for the WIM.
 * @directory:  Path to the directory to create.
 * @parents:	True if parents are to be created as needed, and if no error
 * 		is to be issued if the directory already exists.
 * @return:	True if the directory was successfully created, or if @parents
 * 		was true and the directory already existed.
 */
bool wimlib_mkdir(WIMStruct *w, const char *directory, bool parents)
{
	if (w->current_image == WIM_NO_IMAGE && !wimlib_select_image(w, 1))
		return false;
	wim_set_current_image_modified(w);
	WIMDentry *dentry = get_dentry(w, directory);
	if (dentry != NULL) {
		/* Path exists. Report an error unless it is a directory and the
		 * `-p' option was given. */
		if (parents && dentry_is_directory(dentry)) {
			return true;
		} else {
			wimlib_error("cannot create directory `%s': "
					"File exists\n", directory);
			return false;
		}
	}
	WIMDentry *parent = get_parent_dentry(w, directory);
	if (parent == NULL) {
		/* Parent does not exist.  If the `-p' option was given, make
		 * the missing directories; otherwise, report an error. */
		if (parents) {
			DEBUG("Making multiple directories.\n");
			WIMDentry *new_dir = create_dentry(basename(directory));
			dentry_set_is_directory(new_dir);
			if (is_absolute_path(directory))
				return insert_dentry(wim_root_dentry(w), new_dir,
						directory);
			else
				return insert_dentry(w->working_directory, 
						new_dir, directory);
		} else {
			wimlib_error("cannot create directory `%s': No such file "
					"or directory\n", directory);
			return false;
		}
	} else {
		/* Parent does exist. Create the dentry and link it into the
		 * parent. */
		WIMDentry *new_dir = create_dentry(basename(directory));
		dentry_set_is_directory(new_dir);
		link_dentry(new_dir, parent);
		calculate_dentry_full_path(new_dir);
		return true;
	}
}

/* 
 * Removes a file from the WIM.
 *
 * @w:		Pointer to the WIMStruct for the WIM.
 * @path:	Path to the file.
 * @recursive:	Whether it is OK to remove directories (recursively).
 * @verbose:	Whether each file or directory removed is to be printed.
 *
 * @return:	True if file(s) were successfully removed.
 */
bool wimlib_rm(WIMStruct *w, const char *path, bool recursive, bool verbose)
{
	if (w->current_image == WIM_NO_IMAGE && !wimlib_select_image(w, 1))
		return false;
	WIMDentry *dentry = get_dentry(w, path);
	if (dentry == NULL) {
		wimlib_error("cannot remove `%s': No such file or directory\n", path);
		return false;
	}
	if (!dentry_is_directory(dentry)) {
		if (verbose)
			printf("removed `%s'\n", dentry->fullpath_ascii);
		unlink_dentry(dentry);
		decrement_ref_count(dentry, &w->lookup_table);
		free_dentry(dentry);
	} else {
		if (!recursive) {
			wimlib_error("cannot remove `%s': Is a directory\n", path);
			return false;
		}
		if (dentry_is_root(dentry)) {
			wimlib_error("cannot remove `%s': Is root directory\n", path);
			return false;
		}
		free_dentry_tree(dentry, &w->lookup_table, verbose, true, true);
	}
	wim_set_current_image_modified(w);
	return true;
}
