/*
 * read.c
 *
 * Support for reading WIM files.
 *
 * wimlib - Library for working with WIM files 
 *
 * Copyright (C) 2010 Carl Thijssen
 * Copyright (C) 2012 Eric Biggers
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#include "wimlib_internal.h"
#include "lzx.h"
#include <string.h>
#include <assert.h>
#include <errno.h>

/* Return true if a WIMResourceEntry indicates a compressed resource. */
static inline bool resource_is_compressed(const WIMResourceEntry *entry)
{
	return entry->original_size != entry->size || 
		entry->flags & WIM_RESHDR_FLAG_COMPRESSED;
}

/* 
 * Reads an LZX-compressed file resource.
 *
 * @fp:      		The FILE* for the WIM file.
 * @offset:  		The offset at which the file resource is located.
 * @size:    		The compressed size of the file resource.  
 * @uncompressed_size:  The uncompressed size of the file resource.
 * @uncompressed_data:  An array that must have length of at least
 * 				@uncompressed_size. The uncompressed data 
 * 				is written into it.
 * @return:  		True on success, false on failure.
 */
static bool read_lzx_compressed_file_resource(FILE *fp, u64 offset, u64 size,
		u64 uncompressed_size, u8 uncompressed_data[])
{
	if (fseeko(fp, offset, SEEK_SET) != 0) {
		wimlib_error("Failed to seek to byte %llu to read LZW-compressed "
				"file resource: %s\n", offset, strerror(errno));
		return false;
	}

	/* The structure of a compressed file resource consists of a table of chunk
	 * offsets (one chunk per 32768 bytes of the uncompressed file) followed by the
	 * chunks themselves. The offsets are relative to the start of the file resource
	 * (that is, the parameter @offset). */

	/* Figure out how many chunks are needed for the file. */
	u64 num_chunks = (uncompressed_size + WIM_CHUNK_SIZE - 1) / WIM_CHUNK_SIZE;

	/* There is no chunk entry for the first chunk. */
	u64 num_chunk_entries = num_chunks - 1;

	/* Declare the chunk table, including an entry for the first chunk,
	 * which is not present in the file. */
	u64 chunk_offsets[num_chunks];

	/* Size, in bytes, of the chunk entry table in the file. */
	u64 chunk_entry_table_size;

	/* According to M$'s documentation, if the uncompressed size of the file
	 * is greater than 4 GB, the chunk entries are 8-byte integers.
	 * Otherwise, they are 4-byte integers. */
	if (uncompressed_size >= (u64)1 << 32) {

		/* Read 8-byte entries directly into chunk_offsets. */
		chunk_entry_table_size = num_chunk_entries * sizeof(u64);
		if (fread(chunk_offsets + 1, 1, chunk_entry_table_size, fp)
				!= chunk_entry_table_size)
			goto read_lzx_compressed_file_resource_fread_error;
		for (u64 i = 1; i < num_chunks; i++)
			TO_LE64(chunk_offsets[i]);
	} else {

		/* Read 4-byte entries into a new array chunk_entries, then
		 * expand them into 8-byte entries in chunk_offsets. */
		chunk_entry_table_size = num_chunk_entries * sizeof(u32);
		u32 chunk_entries[num_chunk_entries];
		if (fread(chunk_entries, 1, chunk_entry_table_size, fp)
				!= chunk_entry_table_size)
			goto read_lzx_compressed_file_resource_fread_error;

		for (u64 i = 0; i < num_chunk_entries; i++)
			chunk_offsets[i + 1] = to_le32(chunk_entries[i]);
	}

	if (size < chunk_entry_table_size) {
		wimlib_error("Expected the size of the file resource (%llu bytes) "
			"to be greater than the size of the chunk entry table "
			"(%llu bytes)\n", size, chunk_entry_table_size);
		return false;
	}

	/* Set offset of first chunk. 
	 *
	 * Note: M$'s documentation includes a picture that shows the first
	 * chunk starting right after the chunk entry table, labeled as offset
	 * 0x10.  However, in the actual file format, the offset is measured
	 * from the end of the chunk entry table, so the first chunk has an
	 * offset of 0. */
	chunk_offsets[0] = 0;


	/* Pointer to current position in the uncompressed data. */
	u8 *uncompressed_buf = (u8*)uncompressed_data;

	/* Buffer for compressed data. */
	unsigned char compressed_buf[WIM_CHUNK_SIZE];

	/* Initialize the LZX decompressor. */
	struct LZXstate *lzx_state = LZXinit(15);

	/* Decompress all the chunks. */
	for (u64 i = 0; i < num_chunks; i++) {

		/* Calculate the sizes of the compressed chunk and of the
		 * uncompressed chunk. */
		u64 compressed_chunk_size;
		u64 uncompressed_chunk_size;
		if (i < num_chunks - 1) {
			/* All the chunks except the last one expand to
			 * WIM_CHUNK_SIZE uncompressed, and the amount of
			 * compressed data for the chunk is given by the
			 * difference of offsets in the chunk offset table. */
			compressed_chunk_size = chunk_offsets[i + 1] - 
						chunk_offsets[i];
			uncompressed_chunk_size = WIM_CHUNK_SIZE;
		} else {
			/* The last compressed chunk consists of the remaining
			 * bytes in the file resource, and the last uncompressed
			 * chunk has size equal to however many bytes are left-
			 * that is, the remainder of the uncompressed size when
			 * divided by WIM_CHUNK_SIZE. */
			compressed_chunk_size = size - chunk_offsets[i];

			uncompressed_chunk_size = uncompressed_size % WIM_CHUNK_SIZE;

			/* In an undocumented case, the compressed chunk size of
			 * the last chunk can be greater than the uncompressed
			 * chunk size, and in this case, the data is not
			 * compressed. */
			if (compressed_chunk_size > uncompressed_chunk_size)
				compressed_chunk_size = uncompressed_chunk_size;

		}

		/* This is undocumented, but blocks can be uncompressed.  This
		 * appears to be the case when the compressed chunk size is
		 * WIM_CHUNK_SIZE (32768) for non-last blocks, or when the
		 * compressed chunk size is greater than or equal to the
		 * uncompressed chunk size for the last block. */
		if (compressed_chunk_size == 32768 || (i == num_chunks - 1 && 
				compressed_chunk_size >= uncompressed_chunk_size)) 
		{

			if (fread(uncompressed_buf, 1, uncompressed_chunk_size, 
						fp) != uncompressed_chunk_size)
				goto read_lzx_compressed_file_resource_fread_error;

		} else {

			/* Read the compressed data into compressed_buf. */
			if (fread(compressed_buf, 1, compressed_chunk_size, 
						fp) != compressed_chunk_size)
				goto read_lzx_compressed_file_resource_fread_error;



			/* Decompress the compressed data in compressed_buf, writing the
			 * uncompressed data into uncompressed_buf. */
			int status = LZXdecompress(lzx_state, compressed_buf, 
					uncompressed_buf, compressed_chunk_size, 
					uncompressed_chunk_size);
			if (status != DECR_OK) {
				wimlib_error("Error during LZX decompression.\n");
				switch (status) {
				case DECR_ILLEGALDATA:
					wimlib_error("DECR_ILLEGALDATA\n");
					break;
				case DECR_DATAFORMAT:
					wimlib_error("DECR_DATAFORMAT\n");
					break;
				}
				return false;
			}

			LZXreset(lzx_state);
		}

		/* Advance the pointer into the uncompressed output data by the
		 * number of uncompressed bytes that were written.  */
		uncompressed_buf += uncompressed_chunk_size;
	}

	LZXteardown(lzx_state);

	return true;

read_lzx_compressed_file_resource_fread_error:
	if (feof(fp))
		wimlib_error("Unexpected EOF in LZX-compressed file resource!\n");
	else
		wimlib_error("Error reading LZX-compressed file resource: %s!\n",
				strerror(errno));
	return false;
}

/* 
 * Reads an uncompressed file resource.
 *
 * @fp:      		The FILE* for the WIM file.
 * @offset:  		The offset at which the file resource is located.
 * @size:    		The size of the file resource.  
 * @uncompressed_data:  An array that must have length of at least @size. 
 * 				The uncompressed data is written into it.
 * @return:  		True on success, false on failure.
 */
static bool read_uncompressed_file_resource(FILE *fp, u64 offset, u64 size, 
						u8 uncompressed_data[])
{
	if (fseeko(fp, offset, SEEK_SET) != 0) {
		wimlib_error("Failed to seek to byte %llu to read uncompressed "
				"file resource: %s\n", offset, strerror(errno));
		return false;
	}
	if (fread(uncompressed_data, 1, size, fp) != size) {
		if (feof(fp))
			wimlib_error("Unexpected EOF in uncompressed file resource!\n");
		else
			wimlib_error("Failed to read %llu bytes from uncompressed "
					"file resource!\n", size);
		return false;
	}
	return true;
}

/* 
 * Reads a file resource.
 *
 * @fp:  		The FILE* for the WIM file.
 * @metadata:  		The WIMResourceEntry that describes the file resource.
 * @wim_hdr_flags:  	The flags field that was in the header of the WIM file.
 * @file_contents:  	An array, that must have length at least
 * 				metadata->original_size, into which the 
 * 				uncompressed contents of the file resource 
 * 				will be written.
 * @return:  		True on success, false on failure.
 */
INTERNAL bool read_file_resource(FILE *fp, const WIMResourceEntry *metadata, 
				u32 wim_hdr_flags, u8 file_contents[])
{
	if (metadata->flags & WIM_RESHDR_FLAG_COMPRESSED) {
		return read_lzx_compressed_file_resource(fp, 
				metadata->offset, metadata->size, 
				metadata->original_size, file_contents);
	} else {
		assert(metadata->original_size == metadata->size);
		return read_uncompressed_file_resource(fp, metadata->offset,
			metadata->size, file_contents);
	}
}


/* 
 * Reads the header for a WIM file.
 *
 * @fp:  	The FILE* for the WIM file.
 * @hdr: 	A pointer to a WIMHeader structure into which the information from the
 * 			header will be written.
 * @return:  	True on success, false on failure.
 */
static bool read_header(FILE *fp, WIMHeader *hdr)
{

	static const char expected_magic[WIM_MAGIC_LEN] = {
		'M', 'S', 'W', 'I', 'M', '\0', '\0', '\0' };

	size_t bytes_read;

	fseeko(fp, 0, SEEK_SET);

	bytes_read = fread(hdr->magic, 1, sizeof(hdr->magic), fp);

	if (bytes_read != sizeof(hdr->magic))
		goto read_header_fread_error;

	/* Byte 8 */

	if (memcmp(hdr->magic, expected_magic, 8) != 0) {
		wimlib_error("Not a WIM file (invalid magic characters)!\n");
		return false;
	}

	bytes_read = fread(&hdr->hdr_size, 1, 4, fp);
	if (bytes_read != 4)
		goto read_header_fread_error;

	TO_LE32(hdr->hdr_size);

	/* Byte 12 */

	if (hdr->hdr_size != WIM_HEADER_DISK_LEN) {
		wimlib_error("Unexpected header length of %llu\n", hdr->hdr_size);
		return false;
	}
	{
		u8 buf[hdr->hdr_size];
		bytes_read = fread(buf, 1, hdr->hdr_size - 12, fp);

		if (bytes_read != hdr->hdr_size - 12)
			goto read_header_fread_error;

		const u8 *p = buf;

		p = get_bytes(p, offsetof(WIMHeader, image_count) + sizeof(u32) - 
				offsetof(WIMHeader, version), &hdr->version);
		TO_LE32(hdr->version);
		TO_LE32(hdr->flags);
		TO_LE32(hdr->chunk_size);
		TO_LE16(hdr->part_number);
		TO_LE16(hdr->total_parts);
		TO_LE32(hdr->image_count);


		/* Byte 48 */

		p = get_resource_entry(p, &hdr->lookup_table_metadata);
		p = get_resource_entry(p, &hdr->xml_metadata);
		p = get_resource_entry(p, &hdr->boot_metametadata);

		/* Byte 120 */

		p = get_u32(p, &hdr->boot_idx);

		/* Byte 124 */

		p = get_resource_entry(p, &hdr->integrity);

		/* 60 bytes of unused stuff.  I have no idea if it is secretly
		 * used for something or not, but it's best just to read it in
		 * so it can be copied to the output file.  */
		p = get_bytes(p, WIM_UNUSED_LEN, hdr->unused);
	}

	/* Byte 148 */

	/* 60 bytes of unused fields.  We don't bother reading them into the
	 * header structure.  Although, probably M$ secretly puts extra data in
	 * there.  */

	if (hdr->hdr_size != WIM_HEADER_DISK_LEN) {
		wimlib_error("Warning: WIM Header is %llu bytes, but "
				"expected %llu bytes\n", hdr->hdr_size,
				WIM_HEADER_DISK_LEN);
	}

	return true;
read_header_fread_error:
	if (feof(fp))
		wimlib_error("Unexpected EOF while reading WIM header!\n");
	else
		wimlib_error("File stream error while reading WIM header!\n");
	return false;
}

/*
 * Reads the lookup table from a WIM file.
 *
 * @fp:  		The FILE* for the WIM file.
 * @offset:  		The offset of the lookup table resource.
 * @size:  		The size of the lookup table resource.
 * @lookup_table_ret:  	A pointer to a WIMLookupTable structure into which the
 * 				lookup table will be returned.
 * @return:		True on success, false on failure.
 */
static bool read_lookup_table(FILE *fp, u64 offset, u64 size, 
		WIMLookupTable *lookup_table_ret)
{
	DEBUG("Reading lookup table: offset %llu, size %llu\n",
			offset, size);
	if (fseeko(fp, offset, SEEK_SET) != 0) {
		wimlib_error("Failed to seek to byte %llu to read lookup table: "
				"%s\n", offset, strerror(errno));
		return false;
	}

	size_t num_entries = size / WIM_LOOKUP_TABLE_ENTRY_DISK_SIZE;
	size_t capacity = num_entries * 2 - 1;
	lookup_table_init(lookup_table_ret, capacity);

	u8 buf[WIM_LOOKUP_TABLE_ENTRY_DISK_SIZE];

	while (num_entries--) {
		if (fread(buf, 1, sizeof(buf), fp) != sizeof(buf)) {
			if (feof(fp))
				wimlib_error("Unexpected EOF in lookup table!\n");
			else
				wimlib_error("Stream read error while reading "
						"lookup table!\n");
			lookup_table_destroy(lookup_table_ret);
			return false;
		}
		const u8 *p = buf;
		WIMLookupTableEntry *cur_entry = new_lookup_table_entry();
		p = get_resource_entry(p, &cur_entry->resource_entry);
		p = get_u16(p, &cur_entry->part_number);
		p = get_u32(p, &cur_entry->refcnt);
		p = get_bytes(p, sizeof(cur_entry->hash), cur_entry->hash);

		lookup_table_insert(lookup_table_ret, cur_entry);
	}
	DEBUG("Done reading lookup table.\n");
	return true;
}

/* 
 * Reads the XML data from the WIM file.
 *
 * @fp:		The FILE* for the WIM file.
 * @res:	A pointer to the resource entry for the XML data.
 * @return:	A pointer to the XML data, or NULL on failure.
 */
INTERNAL u8 *read_xml_data(FILE *fp, WIMResourceEntry *res)
{
	if (resource_is_compressed(res)) {
		wimlib_error("XML data is supposed to be uncompressed!\n");
		return false;
	}
	if (res->size < 2) {
		wimlib_error("XML data must be at least 2 bytes!\n");
		return false;
	}

	u8 *xml_data = xmalloc(res->size + 1);
	if (!read_uncompressed_file_resource(fp, res->offset, res->size, xml_data)) {
		free(xml_data);
		return NULL;
	}
	xml_data[res->size] = '\0';
	return xml_data;
}

/* 
 * Reads the security data from the metadata resource.
 *
 * @metadata_resource:	An array that contains the uncompressed metadata
 * 				resource for the WIM file.
 * @metadata_resource_len:	The length of @metadata_resource.
 * @sd:		A pointer to a WIMSecurityData structure that is filled in with
 * 		the security data.
 * @return:	True on success, false on failure.
 *
 * Note: There is no `offset' argument because the security data is located at
 * the beginning of the metadata resource.
 */
static bool read_security_data(const u8 metadata_resource[], 
		u64 metadata_resource_len, WIMSecurityData *sd)
{
	if (metadata_resource_len < 8) {
		wimlib_error("Not enough space in %llu-byte file resource for "
				"security data!\n", metadata_resource_len);
		return false;
	}
	const u8 *p = metadata_resource;
	p = get_u32(p, &sd->total_length);
	p = get_u32(p, &sd->num_entries);

	/* Verify the listed total length of the security data is big enough to
	 * include the sizes array, verify that the file data is big enough to
	 * include it as well, then allocate the array of sizes. */
	u32 num_entries = sd->num_entries;
	u64 sizes_size = num_entries * sizeof(u64);
	u64 size_no_descriptors = 8 + sizes_size;
	if (size_no_descriptors > sd->total_length) {
		wimlib_error("Security data total length of %llu is too short "
				"because there must be at least %llu bytes of security "
				"data!\n", sd->total_length, 
				8 + sizes_size);
		return false;
	}
	if (size_no_descriptors > metadata_resource_len) {
		wimlib_error("File resource of %llu bytes is not big enough "
				"to hold security data of at least %llu "
				"bytes!\n", metadata_resource_len, size_no_descriptors);
		return false;
	}
	u64 *sizes = xmalloc(sizes_size);
	sd->sizes = sizes;

	/* Copy the sizes array in from the file data. */
	p = get_bytes(p, sizes_size, sd->sizes);
	array_to_le64(sd->sizes, num_entries);

	/* Allocate the array of pointers to descriptors, and read them in. */
	u8 **descriptors = xmalloc(num_entries * sizeof(u8*));
	sd->descriptors = descriptors;
	u64 total_len = size_no_descriptors;
	while (num_entries--) {
		total_len += *sizes;
		if (total_len > sd->total_length) {
			wimlib_error("Security data total length of %llu is too "
					"short because there are at least %llu "
					"bytes of security data!\n", 
					sd->total_length, total_len);
			free(sd->sizes);
			free(sd->descriptors);
			return false;
		}
		if (total_len > metadata_resource_len) {
			free(sd->sizes);
			free(sd->descriptors);
			wimlib_error("File resource of %llu bytes is not big enough "
					"to hold security data of at least %llu "
					"bytes!\n", metadata_resource_len, total_len);
			return false;
		}
		*descriptors = xmalloc(*sizes);
		p = get_bytes(p, *sizes, *descriptors);
		sizes++;
		descriptors++;
	}

	/* The total_length field seems to take into account padding for
	 * quadword alignment of the dentry following it, so we can ignore the
	 * case where the actual length read so far is less than the specified
	 * total length of the security data. */
	#if 0
	if (total_len < sd->total_length) {
		/*wimlib_error("Warning: security data was actually %llu bytes, but "*/
				/*"it says its length is %llu bytes!\n",*/
				/*total_len, sd->total_length);*/
	}
	#endif

	return true;
}

/* 
 * Reads a directory entry from the metadata resource.
 *
 * @metadata_resource:	An array that contains the uncompressed metadata
 * 				resource for the WIM file.
 * @metadata_resource_len:	The length of @metadata_resource.
 * @offset:	The offset of the directory entry in the metadata resource.
 * @dentry:	A pointer to a WIMDentry into which the directory entry that is
 * 		read will be returned.
 * @return:	True on success, false on failure.
 */
static bool read_dentry(const u8 metadata_resource[], u64 metadata_resource_len, 
					u64 offset, WIMDentry *dentry)
{
	dentry->file_name        = NULL;
	dentry->short_name       = NULL;
	dentry->stream_entries   = NULL;
	dentry->file_name_ascii  = NULL;
	dentry->short_name_ascii = NULL;
	dentry->rest             = NULL;
	dentry->fullpath_ascii   = NULL;
	dentry->parent           = NULL;
	dentry->children         = NULL;

	/*Make sure the dentry really fits into the metadata resource.*/
	if (offset + 8 > metadata_resource_len) {
		wimlib_error("Directory entry starting at %llu ends past the "
				"end of the metadata resource (size %llu)!\n",
				offset, metadata_resource_len);
		return false;
	}

	/* Before reading the whole entry, we need to read just the length.
	 * This is because an entry of length 8 (that is, just the length field)
	 * terminates the list of sibling directory entries. */

	const u8 *p = &metadata_resource[offset];

	p = get_u64(p, &dentry->length);

	/* A zero length field (really a length of 8, since that's how big the
	 * directory entry is...) indicates that this is the end of directory
	 * dentry.  We do not read it into memory as an actual dentry, so just
	 * return true in that case. */
	if (dentry->length == WIM_END_DIRECTORY_LEN) {
		/*DEBUG("End of directory, offset %lu\n", offset);*/
		return true;
	}

	if (offset + dentry->length >= metadata_resource_len) {
		wimlib_error("Directory entry at offset %llu and with size "
				"%llu ends past the end of the metadata resource "
				"(size %llu)!\n", offset, dentry->length,
				metadata_resource_len);
		return false;
	}

	/* If it is a recognized length, read the rest of the directory entry.
	 * Note: The root directory entry has no name, and its length does not
	 * include the short name length field.  */
	if (dentry->length < WIM_DENTRY_DISK_SIZE)
	{
		wimlib_error("Directory entry has invalid length of "
				"%llu bytes\n", dentry->length);
		return false;
	}
	p = get_bytes(p, WIM_DENTRY_DISK_SIZE - 8, &dentry->attributes);


	TO_LE32(dentry->attributes);
	TO_LE32(dentry->security_id);
	TO_LE64(dentry->subdir_offset);

	TO_LE64(dentry->unused1);
	TO_LE64(dentry->unused2);
	TO_LE64(dentry->creation_time);
	TO_LE64(dentry->last_access_time);
	TO_LE64(dentry->last_write_time);
	/* Hash is a byte field. */
	TO_LE32(dentry->reparse_tag);

	/* reparse_reserved does not exist.  See comment in wim.h */
	/*TO_LE32(dentry->reparse_reserved);*/
	TO_LE64(dentry->hard_link);
	TO_LE16(dentry->streams);
	TO_LE16(dentry->short_name_len);
	TO_LE16(dentry->file_name_len);

	u64 calculated_size = dentry->file_name_len + 
		dentry->short_name_len + WIM_DENTRY_DISK_SIZE + 
			WIM_FILENAME_PADDING;

	if (dentry->length < calculated_size) {
		wimlib_error("Unexpected end of directory entry! (Expected "
				"%llu bytes, got %llu bytes. "
				"short_name_len = %lu, file_name_len = %lu)\n", 
				calculated_size, dentry->length,
				dentry->short_name_len, dentry->file_name_len);
		return false;
	}

	/* Read the filename. */
	dentry->file_name = xmalloc(dentry->file_name_len);
	p = get_bytes(p, dentry->file_name_len, dentry->file_name);

	/* Convert filename to ASCII. */
	size_t ascii_len;
	dentry->file_name_ascii = wide_to_ascii(dentry->file_name, 
					dentry->file_name_len, &ascii_len);
	dentry->file_name_ascii_len = ascii_len;

	/*DEBUG("%s, offset = %lu\n", dentry->file_name_ascii, offset);*/

	/* Undocumented padding between file name and short name. */
	p += WIM_FILENAME_PADDING;

	/* Read the short filename. */
	dentry->short_name = xmalloc(dentry->short_name_len);
	p = get_bytes(p, dentry->short_name_len, dentry->short_name);

	/* Convert short filename to ASCII. */
	dentry->short_name_ascii = wide_to_ascii(dentry->short_name, 
						dentry->short_name_len,
						&ascii_len);
	dentry->short_name_ascii_len = ascii_len;

	/* Handle any stream entries that follow this dir entry. */
	u16 num_streams = dentry->streams;
	u64 num_allocated_stream_names = 0;
	if (num_streams != 0) {
		dentry->stream_entries = xmalloc(num_streams * sizeof(WIMStreamEntry));
		WIMStreamEntry *cur_entry = dentry->stream_entries;
		while (num_streams--) {
			/* Read the base stream entry, excluding the stream name. */
			calculated_size += WIM_STREAM_ENTRY_DISK_SIZE;
			if (calculated_size > dentry->length) {
				wimlib_error("Stream entries go past end of "
						"directory entry (%llu streams "
						"left to read)!\n", num_streams + 1);
				goto read_dentry_error;
			}

			p = get_bytes(p, WIM_STREAM_ENTRY_DISK_SIZE, cur_entry);

			TO_LE64(cur_entry->length);
			TO_LE64(cur_entry->unused1);
			TO_LE16(cur_entry->stream_name_len);

			/* Read the name of the stream. */
			calculated_size += cur_entry->stream_name_len;

			if (calculated_size > dentry->length) {
				wimlib_error("Stream entries go past end of "
						"directory entry (%llu streams "
						"left to read)!\n", num_streams + 1);
				goto read_dentry_error;
			}

			cur_entry->stream_name = xmalloc(cur_entry->stream_name_len);
			num_allocated_stream_names++;
			p = get_bytes(p, cur_entry->stream_name_len,
							cur_entry->stream_name);
			cur_entry++;
		}
	}

	/* There may be extra bytes at the end of the directory entry.  I
	 * believe these are only used for padding so the directory entry length
	 * is a multiple of 8, and in that case it would be fine to ignore them.
	 * But just in case, save them for later so they can be written out
	 * exactly as we found them.  */
	if (calculated_size < dentry->length) {
		dentry->restsize = dentry->length - calculated_size;
		dentry->rest = xmalloc(dentry->restsize);
		p = get_bytes(p, dentry->restsize, dentry->rest);
	} else {
		dentry->restsize = 0;
	}

	return true;

read_dentry_error:
	dentry->streams = num_allocated_stream_names;
	destroy_dentry(dentry);
	return false;
}

/* Reads the children of a dentry, and all their children, ..., etc. from the
 * metadata resource and into the dentry tree.
 * @metadata_resource:	An array that contains the uncompressed metadata
 * 				resource for the WIM file.
 * @metadata_resource_len:	The length of @metadata_resource.
 * @dentry:	A pointer to a WIMDentry that is the root of the directory tree
 * 		and has already been read from the metadata resource.  It does not 
 * 		need to be the real root, because this procedure is called 
 * 		recursively.
 * @return:	True on success, false on failure. 
 */
static bool read_dentry_tree(const u8 metadata_resource[], 
		u64 metadata_resource_len, WIMDentry *dentry)
{
	/* Offset of first child of @dentry. */
	u64 cur_offset = dentry->subdir_offset;

	/* If @dentry is a regular file, nothing more needs to be done for this
	 * branch. */
	if (cur_offset == 0)
		return true;

	WIMDentry  cur_child;
	WIMDentry *prev_child = NULL;
	WIMDentry *first_child = NULL;

	/* Find and read all the children of @dentry. */
	while (1) {

		/* Read next child of @dentry into @cur_child. */
		if (!read_dentry(metadata_resource, metadata_resource_len, 
					cur_offset, &cur_child))
			return false;

		/* Check for end of directory. */
		if (cur_child.length == WIM_END_DIRECTORY_LEN) {
			/* Link last child to first one, and set parent's
			 * children pointer to the first child.  */
			if (prev_child != NULL) {
				prev_child->next = first_child;
				first_child->prev = prev_child;
			}
			dentry->children = first_child;
			return true;
		}

		/* Not end of directory.  Allocate this child permanently and
		 * link it to the parent and previous child. */
		WIMDentry *child = xmalloc(sizeof(WIMDentry));
		memcpy(child, &cur_child, sizeof(WIMDentry));

		if (prev_child == NULL) {
			first_child = child;
		} else {
			prev_child->next = child;
			child->prev = prev_child;
		}
		prev_child = child;

		child->parent = dentry;

		/* If there are children of this child, call this procedure
		 * recursively. */
		if (child->subdir_offset != 0) {
			if (!read_dentry_tree(metadata_resource, 
					metadata_resource_len, child))
				return false;
		}
		/* Advance to the offset of the next child. */
		cur_offset += child->length;
	}
	/* Not reached. */
}


/* 
 * Reads the boot metadata from the WIM file.  The boot metadata doesn't really
 * have much to do with booting at all; it is also just called the 'metadata
 * resource'.  The metadata resource consists of the security data, followed by
 * the directory entry for the root directory, followed by all the other
 * directory entries in the filesystem.  The subdir_offset field of each
 * directory entry gives the start of its child entries from the beginning of
 * the metadata resource.  An end-of-directory is signaled by a directory entry
 * of length '0', really of length 8, because that's how long the 'length' field
 * is.
 *
 * @fp:		The FILE* for the input WIM file.
 * @metadata:	The resource entry for the metadata resource (a.k.a the metadata
 * 			for the metadata)
 * @wim_hdr_flags:	The flags as found in the WIM header.
 * @sd:		A pointer to a WIMSecurityData structure into which the security
 * 		data for the WIM file is returuned.
 * @dentry:	A pointer to a WIMDentry structure into which the root dentry is
 * 		returned.
 *
 * @return:	True on success, false on failure.
 */
INTERNAL bool read_boot_metadata(FILE *fp, const WIMResourceEntry *metadata,
		u32 wim_hdr_flags, WIMSecurityData *sd, WIMDentry *root_dentry)
{
	u64 metadata_resource_len = metadata->original_size;
	u8 *metadata_resource = xmalloc(metadata_resource_len);

	DEBUG("Reading metadata resource: length = %lu, offset = %lu\n",
			metadata_resource_len, metadata->offset);

	/* Read the metadata resource into memory.  (It may be compressed.) */
	if (!read_file_resource(fp, metadata, wim_hdr_flags, metadata_resource))
		goto read_boot_metadata_error;

	DEBUG("Finished reading metadata resource into memory.\n");

	/* Read the security data into a WIMSecurityData structure. */
	if (!read_security_data(metadata_resource, metadata_resource_len, sd))
		goto read_boot_metadata_error;

	/* Read the root directory entry into a WIMDentry structure.  It
	 * starts after security data, on a quad-word aligned address. */
	u64 dentry_offset = sd->total_length;
	dentry_offset += (8 - dentry_offset % 8) % 8;
	if (!read_dentry(metadata_resource, metadata_resource_len, 
				dentry_offset, root_dentry))
		goto read_boot_metadata_error;

	root_dentry->parent = root_dentry;

	/* Now read the entire directory entry tree. */
	if (!read_dentry_tree(metadata_resource, metadata_resource_len, root_dentry))
		goto read_boot_metadata_error;

	for_dentry_in_tree(root_dentry, do_calculate_dentry_full_path, NULL);

	free(metadata_resource);

	return true;
read_boot_metadata_error:
	free(metadata_resource);
	return false;
}

/* 
 * If @entry points to a metadata resource, append it to the list of metadata
 * resources in the WIMStruct.
 */
static bool append_boot_metametadata(WIMLookupTableEntry *entry, void *wim_struct_p)
{
	WIMStruct *w = (WIMStruct*)wim_struct_p;
	if (entry->resource_entry.flags & WIM_RESHDR_FLAG_METADATA) {
		if (w->current_image >= w->hdr.image_count) {
			wimlib_error("Expected only %u images, but found more!\n",
					w->hdr.image_count);
			return false;
		} else {
			DEBUG("Found metadata resource for image %u\n",
					w->current_image + 1);
			w->boot_metadata[w->current_image++].lookup_table_entry = entry;
			return true;
		}
	} else {
		/* Do nothing if not a metadata resource. */
		return true;
	}
}

/* 
 * Begins the reading of a WIM file; opens the file and reads its header.
 *
 * @w:  A pointer to a WIMStruct that keeps track of information about the file.
 * @in_wim_path: The name of the file to open.
 * @return: True if no error occurred; false if error occurred.
 */
bool wimlib_begin_read(WIMStruct *w, const char *in_wim_path)
{
	w->filename           = xstrdup(in_wim_path);
	w->fp                 = fopen(in_wim_path, "rb");
	if (w->fp == NULL) {
		wimlib_error("Failed to open the file \"%s\" for reading: %s\n",
				in_wim_path, strerror(errno));
		return false;
	}
	if (!read_header(w->fp, &w->hdr))
		return false;

	DEBUG("Wim file contains %u images\n", w->hdr.image_count);

	u32 flags = w->hdr.flags;
	if ((flags & WIM_HDR_FLAG_COMPRESSION)  &&
			!(flags & WIM_HDR_FLAG_COMPRESS_LZX)) {
		wimlib_error("Unsupported compression method: ");
		if (flags & WIM_HDR_FLAG_COMPRESS_XPRESS)
			wimlib_error("XPRESS\n");
		else if (flags & WIM_HDR_FLAG_COMPRESS_RESERVED)
			wimlib_error("Unknown\n");
		else
			wimlib_error("Unspecified\n");
		return false;
			
	}
	if (w->hdr.lookup_table_metadata.flags & WIM_RESHDR_FLAG_COMPRESSED) {
		wimlib_error("Didn't expect a compressed lookup table!\n");
		return false;
	}
	if (!read_lookup_table(w->fp, w->hdr.lookup_table_metadata.offset,
			w->hdr.lookup_table_metadata.size, &w->lookup_table))
		return false;

	DEBUG("Allocating %u root dentries and %u security data structures\n",
			w->hdr.image_count, w->hdr.image_count);

	w->boot_metadata      = xmalloc(w->hdr.image_count * sizeof(WIMBootMetadata));
	w->working_directory  = &w->boot_metadata[0].root_dentry;
	w->current_image      = 0;

	for (uint i = 0; i < w->hdr.image_count; i++) {
		w->boot_metadata[i].modified = false;
		w->boot_metadata[i].written  = true;
	}

	/* Find the images in the WIM by searching the lookup table. */
	if (!for_lookup_table_entry(&w->lookup_table, 
				append_boot_metametadata, w))
		return false;
	if (w->current_image != w->hdr.image_count) {
		wimlib_error("Only found %u images in WIM, but expected %u!\n",
				w->current_image, w->hdr.image_count);
		return false;
	}
	w->current_image = WIM_NO_IMAGE;
	DEBUG("Done.\n");
	return true;
}
