/*
 * util.c
 *
 * Some utility functions internal to wimlib, but not specific to the structures
 * declared in wimlib_internal.h.  The wide-to-ascii and ascii-to-wide
 * conversion functions are based on code from the chntpw project.
 *
 * wimlib - Library for working with WIM files 
 *
 * Copyright (C) 2010 Carl Thijssen
 * Copyright (C) 2012 Eric Biggers
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#include "util.h"
#include "endianness.h"
#include "sha1.h"


#include <string.h>
#include <assert.h>
#include <ctype.h>
#include <error.h>
#include <errno.h>

INTERNAL void *xmalloc(u64 size)
{
	/* Watch for for casting u64 to size_t.  If size_t is only 32 bits, less
	 * memory will be allocated than we think. */
	if (sizeof(size_t) < sizeof(u64)) {
		if (size > 0xffffffffULL) {
			fprintf(stderr, "Cannot allocate %llu bytes of memory "
					"when size_t is only %u bytes!\n",
					size, sizeof(size_t));
			abort();
		}
	}
	void *p = malloc((size_t)size);
	if (p == NULL) {
		fprintf(stderr, "Out of memory! (Tried to allocate %lu bytes)\n", size);
		abort();
	}
	return p;
}

INTERNAL void *xzalloc(u64 size)
{
	void *p = xmalloc(size);
	return memset(p, 0, size);
}

INTERNAL void *xrealloc(void *ptr, u64 size)
{
	void *p = realloc(p, size);
	if (p == NULL) {
		fprintf(stderr, "Out of memory! (Tried tore allocate %lu bytes)\n", size);
		abort();
	}
	return p;
}


INTERNAL char *xstrdup(const char *s)
{
	char *p = strdup(s);
	if (p == NULL) {
		fprintf(stderr, "strdup(): Out of memory!\n");
		abort();
	}
	return p;
}

// see: http://pogostick.net/~pnh/ntpasswd/
INTERNAL u8 *wide_to_ascii(const u8 *wide_string, size_t size_bytes, 
				size_t *ascii_len_ret)
{
	size_t ascii_len = 0;
	size_t wide_len = size_bytes / 2;
	const u16 *p = (const u16*)wide_string;

	for (size_t j = 0; j < wide_len; j++) {
		unsigned v = to_le16(p[j]);
		if (v < 256)
			ascii_len += 1;
		else if (v < 0x800)
			ascii_len += 2;
		else
			ascii_len += 3;
	}

	u8 *cstring = xmalloc(ascii_len + 1);

	u64 k = 0;
	for (size_t j = 0; j < wide_len; j++) {
		unsigned v = to_le16(p[j]);
		if (v < 256) {
			cstring[k++] = v;
		} else if (v < 0x800) {
			cstring[k++] = 0xc0 | (v >> 6);
			cstring[k++] = 0x80 | (v & 0x3f);
		} else {
			cstring[k++] = 0xe0 | (v >> 12);
			cstring[k++] = 0x80 | ((v >> 6) & 0x3f);
			cstring[k++] = 0x80 | (v & 0x3f);
		}
	}
	cstring[ascii_len] = '\0';
	if (ascii_len_ret)
		*ascii_len_ret = ascii_len;
	return cstring;
}

INTERNAL void wide_to_ascii_print(const u8 *string, u64 len, FILE *file)
{
	for (u64 i = 0; i < len; i += 2) {
		unsigned v = string[i] + string[i + 1] * 512u;
		if (v < 256)
			fputc(v, file);
		else if (v < 0x800) {
			fputc(0xc0 | (v >> 6), file);
			fputc(0x80 | (v & 0x3f), file);
		} else {
			fputc(0xe0 | (v >> 12), file);
			fputc(0x80 | ((v >> 6) & 0x3f), file);
			fputc(0x80 | (v & 0x3f), file);
		}
	}
}

INTERNAL u8 *ascii_to_wide(const u8 *in, int len, int *out_len)
{
	
	size_t size = len * 2 + 2;
	u8 *out = xmalloc(size);
	int j = 0;

	for (int i = 0; i < len; i++) {
		if (in[i] < 0x80) {
			out[j++] = in[i];
			out[j++] = 0;
		} else if ((in[i] & 0xe0) == 0xc0 && i + 1 < len) {
			out[j++] = (in[i] & 0x1f) << 6 | (in[i + 1] & 0x3f);
			out[j++] = (in[i] & 0x1f) >> 2;
			i++;
		} else if (i + 2 < len) {
			/* assume 3 byte */
			out[j++] = (in[i + 1] & 0xf) << 6 | (in[i + 2] & 0x3f);
			out[j++] = (in[i] & 0xf) << 4 | ((in[i + 1] & 0x3f) >> 2);
			i += 2;
		}
	}
	*out_len = j;
	out[j++] = 0;
	out[j++] = 0;
	return (u8*)out;
}

#define intervals_per_second (1000000000ULL / 100ULL)
#define days_per_year (365ULL)
#define seconds_per_day (3600ULL * 24ULL)
#define intervals_per_day (seconds_per_day * intervals_per_second)
#define intervals_per_year (intervals_per_day * days_per_year)
#define years_1601_to_1970 (1970ULL - 1601ULL)
#define leap_years_1601_to_1970 (years_1601_to_1970 / 4ULL - 3ULL)
#define intervals_1601_to_1970 (years_1601_to_1970 * intervals_per_year \
				+ leap_years_1601_to_1970 * intervals_per_day)

/* 
 * Returns the number of 100-nanosecond intervals that have elapsed since
 * 12:00 A.M., January 1, 1601 UTC.
 */
INTERNAL u64 get_timestamp()
{
	return (u64)intervals_1601_to_1970 + (u64)time(NULL) * intervals_per_second;
}

INTERNAL void randomize_byte_array(u8 *p, size_t n)
{
	static bool seeded = false;
	if (!seeded) {
		srand(time(NULL));
		seeded = true;
	}
	while (n--)
		*p++ = rand();
}

/* Takes in a path of length @len in @buf, and transforms it into a string for
 * the path of its parent directory. */
INTERNAL void to_parent_name(char buf[], int len)
{
	int i = len - 1;
	while (i >= 0 && buf[i] == '/')
		i--;
	while (i >= 0 && buf[i] != '/')
		i--;
	while (i >= 0 && buf[i] == '/')
		i--;
	buf[i + 1] = '\0';
}

/* Like basename(), but does not modify path; it just returns a pointer to it.
 * */
INTERNAL const char *path_basename(const char *path)
{
	long i = strlen(path) - 1;
	while (i >= 0 && path[i] == '/')
		i--;
	while (i >= 0 && path[i] != '/')
		i--;
	return path + i + 1;
}

/* 
 * Splits a file path into the part before the first '/', or the entire name if
 * there is no '/', and the part after the first sequence of '/' characters.
 *
 * @path:  		The file path to split.
 * @first_part_len_ret: A pointer to an `int' into which the length of the
 * 				first part of the path will be returned.
 * @return:  		A pointer to the next part of the path, after the first
 * 				sequence of '/', or a pointer to the terminating 
 * 				null byte in the case of a path without any '/'.
 */
INTERNAL const char *path_next_part(const char *path, 
					int *first_part_len_ret)
{
	int i = 0;
	while (path[i] != '/' && path[i] != '\0')
		i++;
	*first_part_len_ret = i;
	const char *next_part = &path[i];
	while (*next_part == '/')
		next_part++;
	return next_part;
}

/* Returns the number of components of @path.  */
INTERNAL int get_num_path_components(const char *path)
{
	int num_components = 0;
	do {
		if (*path == '/') {
			num_components++;
			while (*++path == '/');
		}
	} while (*path++);
	return num_components;
}


/* 
 * Prints a string.  Printable characters are printed as-is, while unprintable
 * characters are printed as their octal escape codes. 
 *
 * @string: 	Pointer to the string.
 * @len: 	Length of the string. 
 */
INTERNAL void print_string(const u8 string[], u64 len)
{
	while (len--) {
		if (isprint(*string))
			putchar(*string++);
		else
			printf("\\%03hho", *string++);
	}
}

INTERNAL bool sha1sum(const char *filename, void *buf)
{
	FILE *fp = fopen(filename, "rb");
	if (fp == NULL)
		return false;
	return sha1_stream(fp, buf) == 0;
}
