/*
 * wim.c
 *
 * Main utility routines for wimlib.
 *
 * wimlib - Library for working with WIM files 
 *
 * Copyright (C) 2010 Carl Thijssen
 * Copyright (C) 2012 Eric Biggers
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#include "wimlib_internal.h"
#include <errno.h>
#include <stdarg.h>
#include <assert.h>

/* True if WIMLIB is to print an informational message when an error occurs.
 * This can be turned off by calling wimlib_set_no_error_messages(). */
static bool print_errors = true;

/* Print formatted error message to stderr. */
INTERNAL void wimlib_error(const char *format, ...)
{
	if (print_errors) {
		va_list va;
		va_start(va, format);
		fputs("WIMLIB: ", stderr);
		vfprintf(stderr, format, va);
		va_end(va);
	}
}

/* Suppress error messages.  Applies to the entire library.  */
void wimlib_set_no_error_messages()
{
	print_errors = false;
}


/* Reads the contents of a WIMResourceEntry, as represented in the on-disk
 * format, from the memory pointed to by @p, and fills in the fields of @entry.
 * A pointer to the byte after the memory read at @p is returned. */
INTERNAL const u8 *get_resource_entry(const u8 *p, WIMResourceEntry *entry)
{
	p = get_u56(p, &entry->size);
	p = get_u8(p, &entry->flags);
	p = get_u64(p, &entry->offset);
	p = get_u64(p, &entry->original_size);
	return p;
}

/* Copies the WIMResourceEntry @entry to the memory pointed to by @p in the
 * on-disk format.  A pointer to the byte after the memory written at @p is
 * returned. */
INTERNAL u8 *put_resource_entry(u8 *p, const WIMResourceEntry *entry)
{
	p = put_u56(p, entry->size);
	p = put_u8(p, entry->flags);
	p = put_u64(p, entry->offset);
	p = put_u64(p, entry->original_size);
	return p;
}



/* 
 * Prints information from a WIM header.
 *
 * @hdr:  A pointer to a WIMHeader structure.
 */
static void print_wim_header(const WIMHeader *hdr)
{
	puts("[WIM HEADER]");
	fputs ("Imagetag                    = ", stdout);
	print_magic(hdr->magic);
	putchar('\n');
	printf("Header Size                 = %u\n", hdr->hdr_size);
	printf("Version                     = 0x%x\n", hdr->version);
	printf("Flags                       = 0x%x\n", hdr->flags);

	if (hdr->flags & WIM_HDR_FLAG_COMPRESS_XPRESS)
		printf("    COMPRESS XPRESS FLAG is set\n");

	if (hdr->flags & WIM_HDR_FLAG_COMPRESS_LZX)
		printf("    COMPRESS LZX FLAG is set\n");

	printf("Chunk Size                  = %u\n", hdr->chunk_size);
	fputs ("GUID                        = ", stdout);
	print_guid(hdr->guid);
	putchar('\n');
	printf("Part Number                 = %hu\n", hdr->part_number);
	printf("Total Parts                 = %hu\n", hdr->total_parts);
	printf("Image Count                 = %u\n", hdr->image_count);
	printf("Lookup Table Size           = %llu\n", hdr->lookup_table_metadata.size);
	printf("Lookup Table Flags          = 0x%hhx\n", hdr->lookup_table_metadata.flags);
	printf("Lookup Table Offset         = %llu\n", hdr->lookup_table_metadata.offset);
	printf("Lookup Table Original_size  = %llu\n", hdr->lookup_table_metadata.original_size);
	printf("XML Data Size               = %llu\n", hdr->xml_metadata.size);
	printf("XML Data Flags              = 0x%hhx\n", hdr->xml_metadata.flags);
	printf("XML Data Offset             = %llu\n", hdr->xml_metadata.offset);
	printf("XML Data Original Size      = %llu\n", hdr->xml_metadata.original_size);
	printf("Boot Metadata Size          = %llu\n", hdr->boot_metametadata.size);
	printf("Boot Metadata Flags         = 0x%hhx\n", hdr->boot_metametadata.flags);
	printf("Boot Metadata Offset        = %llu\n", hdr->boot_metametadata.offset);
	printf("Boot Metadata Original Size = %llu\n", hdr->boot_metametadata.original_size);
	printf("Boot Index                  = %u\n", hdr->boot_idx);
	printf("Integrity Size              = %llu\n", hdr->integrity.size);
	printf("Integrity Flags             = 0x%hhx\n", hdr->integrity.flags);
	printf("Integrity Offset            = %llu\n", hdr->integrity.offset);
	printf("Integrity Original_size     = %llu\n", hdr->integrity.original_size);
}

/* 
 * Prints the XML data for the WIM file.
 *
 * @xml_data:	The XML data.
 * @size:	The size of the XML data.
 */
static void print_xml_data(const u8 *xml_data, u64 size)
{
	puts("[XML DATA]");
	wide_to_ascii_print(xml_data + 2, size - 2, stdout);
	putchar('\n');
}



/* 
 * Prints the security data for a WIM file.
 *
 * @sd:	A pointer to the WIMSecurityData structure.
 */
static void print_security_data(const WIMSecurityData *sd)
{
	puts("[SECURITY DATA]");
	printf("Length            = %u bytes\n", sd->total_length);
	printf("Number of Entries = %u\n", sd->num_entries);

	u64 num_entries = (u64)sd->num_entries;
	for (u64 i = 0; i < num_entries; i++) {
		printf("Entry %llu has size %llu bytes and descriptor ", 
				i, sd->sizes[i]);

		print_byte_field(sd->descriptors[i], sd->sizes[i]);
		putchar('\n');
	}
	putchar('\n');
}

static void destroy_security_data(WIMSecurityData *sd)
{
	u8 **descriptors = sd->descriptors;
	u32 num_entries = sd->num_entries;
	while (num_entries--)
		free(*descriptors++);
	free(sd->sizes);
	free(sd->descriptors);
}

/* 
 * Open a file and determine its length.
 *
 * @path:  The path to the file.
 * @mode:  The mode to open it with.
 * @len_ret:  A pointer to an off_t in which the size of the file will be
 * 	returned.
 * @return:  A FILE* to the file, or NULL if an error occurred.
 */
INTERNAL FILE *fopen_and_get_length(const char *path, const char *mode, 
							off_t *len_ret)
{
	FILE *fp = fopen(path, mode);
	if (fp == NULL) {
		wimlib_error("Failed to open the file `%s' with mode `%s': %s\n",
				path, mode, strerror(errno));
		return NULL;
	}
	if (fseeko(fp, 0, SEEK_END) != 0) {
		wimlib_error("Failed to seek to end of file `%s': %s\n",
				path, strerror(errno));
		fclose(fp);
		return NULL;
	}
	if ((*len_ret = ftello(fp)) == -1) {
		wimlib_error("Failed to get length of file `%s': %s\n",
				path, strerror(errno));
		fclose(fp);
		return NULL;
	}
	if (fseeko(fp, 0, SEEK_SET) != 0) {
		wimlib_error("Failed to seek to beginning of file `%s': %s\n",
				path, strerror(errno));
		fclose(fp);
		return NULL;
	}
	return fp;
}

/*
 * Returns a pointer to a newly allocated WIMStruct.  It must be freed using
 * wimlib_destroy(). 
 */
WIMStruct *wimlib_create()
{
	WIMStruct *w = xmalloc(sizeof(WIMStruct));
	w->lookup_table.array = NULL;
	w->xml_data           = NULL;
	w->output_dir         = NULL;
	w->boot_metadata      = NULL;
	w->filename           = NULL;
	w->fp                 = NULL;
	w->print_files        = false;
	w->link_type          = WIM_LINK_TYPE_HARD;
	w->current_image      = WIM_NO_IMAGE;
	return w;
}


char *wimlib_get_working_directory_fullpath(WIMStruct *w)
{
	return w->working_directory->fullpath_ascii;
}

char *wimlib_get_filename(WIMStruct *w)
{
	return w->filename;
}

unsigned wimlib_get_num_images(const WIMStruct *w)
{
	return w->hdr.image_count;
}

unsigned wimlib_get_current_image(const WIMStruct *w)
{
	return w->current_image + 1;
}

bool wimlib_select_image(WIMStruct *w, unsigned image)
{
	image--;

	if (image == w->current_image)
		return true;
	if (image >= w->hdr.image_count) {
		wimlib_error("Cannot select image %u: There are only %u images!\n",
				image + 1, w->hdr.image_count);
		return false;
	}
	if (wim_current_image_exists(w) && !wim_current_image_is_modified(w)) {
		free_dentry_tree(wim_root_dentry(w), NULL, false, false, false);
		destroy_security_data(wim_security_data(w));
		wim_set_current_image_unmodified(w);
	}
	w->current_image = image;
	bool ret = true;
	if (!wim_current_image_is_modified(w))
		ret = read_boot_metadata(w->fp, wim_metadata_resource_entry(w),
				w->hdr.flags, wim_security_data(w), wim_root_dentry(w));
	if (ret)
		w->working_directory = wim_root_dentry(w);
	return ret;
}

/* Returns true if any image in the WIM has been modified. */
bool wimlib_all_images_written(WIMStruct *w)
{
	for (unsigned i = 0; i < w->hdr.image_count; i++)
		if (!w->boot_metadata[i].written)
			return false;
	return true;
}



/* Sets a flag indicating that the names of extracted or written files are to be
 * printed. Applies to the WIM file associated with @w. */
void wimlib_set_print_files(WIMStruct *w)
{
	w->print_files = true;
}


/* Prints information from the header of the WIM file associated with @w.
 * Return true on success, false on failure. */
bool wimlib_print_wim_header(WIMStruct *w)
{
	print_wim_header(&w->hdr);
	return true;
}

/* Print the XML data from the WIM file associated with @w. 
 * Return true on success, false on failure. */
bool wimlib_print_xml_data(WIMStruct *w)
{
	if (w->xml_data == NULL)
		if ((w->xml_data = read_xml_data(w->fp, &w->hdr.xml_metadata)) == NULL)
			return false;
	print_xml_data(w->xml_data, w->hdr.xml_metadata.size);
	return true;
}


/* Print the boot metadata from the WIM file associated with @w. 
 * Always returns true. */
bool wimlib_print_boot_metadata(WIMStruct *w)
{
	if (w->current_image == WIM_NO_IMAGE && !wimlib_select_image(w, 1))
		return false;
	print_security_data(wim_security_data(w));
	for_dentry_in_tree(wim_root_dentry(w), print_dentry, &w->lookup_table);
	return true;
}



/* 
 * Prints the full paths of all files in the WIM.  Equivalent to
 * wimlib_ls(w, "/", true).
 */
bool wimlib_print_files(WIMStruct *w)
{
	if (w->current_image == WIM_NO_IMAGE && !wimlib_select_image(w, 1))
		return false;
	return for_dentry_in_tree(wim_root_dentry(w), print_dentry_full_path, 
					NULL);
}


/* 
 * Sends the contents of a file in the WIM to standard output.
 *
 * @w:		Pointer to the WIMStruct for the WIM.
 * @path:	Path to the file.
 * @return:	True if @path was a regular file and was successfully sent
 * 		to stdout.
 */
bool wimlib_cat(WIMStruct *w, const char *path)
{
	if (w->current_image == WIM_NO_IMAGE && !wimlib_select_image(w, 1))
		return false;
	WIMDentry *dentry = get_dentry(w, path);
	if (dentry == NULL) {
		wimlib_error("cat: %s: No such file or directory\n", path);
		return false;
	}
	if (dentry_is_directory(dentry)) {
		wimlib_error("cat: %s: Is a directory\n", path);
		return false;
	}
	WIMLookupTableEntry *entry = lookup_file_resource(dentry, &w->lookup_table);
	if (entry == NULL) {
		wimlib_error("Could not find the file resource for `%s'\n", path);
		return NULL;
	}
	if (entry->file_on_disk != NULL) {
		FILE* file = fopen(entry->file_on_disk, "rb");
		if (file == NULL) {
			wimlib_error("Could not open the file `%s': %s\n",
					entry->file_on_disk, strerror(errno));
			return false;
		}
		int c;
		while ((c = getc(file)) != EOF)
			putchar(c);
		if (ferror(file)) {
			wimlib_error("Error reading the file `%s': %s\n",
					entry->file_on_disk, strerror(errno));
			return false;
		}
	} else {
		size_t size  = entry->resource_entry.original_size;
		void *p = xmalloc(size);
		if (!read_file_resource(w->fp, &entry->resource_entry, w->hdr.flags, p))
			return false;
		fwrite(p, 1, size, stdout);
	}
	return true;
}

/* 
 * Changes the working directory in the WIM file.
 *
 * @w:		Pointer to the WIMStruct for the WIM.
 * @working_directory:	Path to the new working directory.
 * @return:	True if the new working directory was a valid directory.
 */
bool wimlib_cd(WIMStruct *w, const char *working_directory)
{
	if (w->current_image == WIM_NO_IMAGE && !wimlib_select_image(w, 1))
		return false;
	WIMDentry *dentry = get_dentry(w, working_directory);
	if (dentry == NULL) {
		wimlib_error("cd: %s: No such file or directory\n", 
						working_directory);
		return false;
	}
	if (!dentry_is_directory(dentry)) {
		wimlib_error("cd: %s: Not a directory\n", 
						working_directory);
		return false;
	}
	w->working_directory = dentry;
	return true;
}

/* 
 * List files in a directory, possible recursively.
 *
 * @w:		Pointer to the WIMStruct for the WIM.
 * @directory:	Path to the directory to list.  If NULL, the working directory
 * 		will be listed.
 * @recursive:	True if files are to be listed recursively.
 * @return:	True if the files were successfully listed.
 */
bool wimlib_ls(WIMStruct *w, const char *directory, bool recursive)
{
	if (w->current_image == WIM_NO_IMAGE && !wimlib_select_image(w, 1))
		return false;
	WIMDentry *dir;
	if (directory == NULL) {
		dir = w->working_directory;
	} else {
		dir = get_dentry_relative_path(w->working_directory, directory);
		if (dir == NULL) {
			wimlib_error("ls: cannot access `%s': No such file or directory\n",
					directory);
			return false;
		}
	}
	if (recursive) {
		for_dentry_in_tree(dir, print_dentry_full_path, NULL);
	} else {
		WIMDentry *child = dir->children;
		if (child != NULL) {
			do {
				fputs(child->file_name_ascii, stdout);
				if (dentry_is_directory(child))
					putchar('/');
				putchar('\n');
				child = child->next;
			} while (child != dir->children);
		}
	}
	return true;
}



/* 
 * Prints the metadata for the file with @path.
 */
bool wimlib_stat(WIMStruct *w, const char *path)
{
	if (w->current_image == WIM_NO_IMAGE && !wimlib_select_image(w, 1))
		return false;
	WIMDentry *dentry = get_dentry(w, path);
	if (dentry == NULL) {
		wimlib_error("cannot stat `%s': No such file or directory.\n",
				path);
		return false;
	}
	return print_dentry(dentry, &w->lookup_table);
}


/* Returns true if the file specified by the ASCII @path exists and is a
 * directory. */
bool wimlib_is_directory(WIMStruct *w, const char *path)
{
	if (w->current_image == WIM_NO_IMAGE && !wimlib_select_image(w, 1))
		return false;
	WIMDentry *dentry = get_dentry(w, path);
	return dentry != NULL && dentry_is_directory(dentry);
}

/* 
 * Utility function for reading the contents of a file.
 *
 * @filename:	The name of the file to read.
 * @len_ret:	A location to store the length of the file.
 * @return:	A pointer to a dynamically allocated buffer of length
 * 		*len_ret that contains the contents of the file,
 * 		NULL on failure.
 */
void *wimlib_file_get_contents(const char *filename, off_t *len_ret)
{
	off_t size;
	FILE *fp = fopen_and_get_length(filename, "rb", &size);
	if (fp == NULL)
		return NULL;
	char *s = xmalloc(size);
	if (fread(s, 1, size, fp) != size) {
		wimlib_error("Could not read the contents of `%s': %s\n", 
				filename, strerror(errno));
		return NULL;
	}
	fclose(fp);
	*len_ret = size;
	return s;
}

/* Closes a WIM file and frees memory inside the WIMStruct. */
void wimlib_end_read(WIMStruct *w)
{
	if (w->fp)
		fclose(w->fp);
	free(w->filename);
	free(w->output_dir);
	free(w->boot_metadata);
}

/* Frees the memory for the WIMStruct.  wimlib_end_read() must be called first
 * if there was any call to wimlib_begin_read() on the structure. */
void wimlib_destroy(WIMStruct *w)
{
	free(w);
}
