/*
 * wimlib_internal.h
 *
 * Internal header for wimlib.
 *
 * wimlib - Library for working with WIM files 
 *
 * Copyright (C) 2010 Carl Thijssen
 * Copyright (C) 2012 Eric Biggers
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#ifndef _WIMLIB_INTERNAL_H
#define _WIMLIB_INTERNAL_H


#include "util.h"
#include <sys/stat.h>


/* Lengths of array fields of constant length. */
#define WIM_HASH_SIZE  20
#define WIM_MAGIC_LEN  8
#define WIM_GID_LEN    16
#define WIM_UNUSED_LEN 60

/* The value that appears in the `length' field of directory entries on disk
 * when that entry is not a real entry, but rather an end-of-directory marker.
 * */
#define WIM_END_DIRECTORY_LEN 0

/* Length of the WIM header on disk. */
#define WIM_HEADER_DISK_LEN (148 + WIM_UNUSED_LEN)

/* Compressed resources in the WIM are divided into separated compressed chunks
 * of this size. */
#define WIM_CHUNK_SIZE 32768

/* Used to indicate that no WIM image is currently selected. */
#define WIM_NO_IMAGE 0xfffffffe

/* Metadata for a resource in a WIM file. */
typedef struct WIMResourceEntry {
	/* Size, in bytes, of the resource in the WIM file. */
	u64 size;

	/* Bitwise or of one or more of the WIM_RESHDR_FLAG_* flags. */
	u8  flags;

	/* Offset, in bytes, of the resource in the WIM file. */
	u64 offset;

	/* Uncompressed size of the resource in the WIM file.  Is the same as
	 * @size if the resource is uncompressed. */
	u64 original_size;
} WIMResourceEntry;

/* Flags for the `flags' field of the WIMResourceEntry structure. */

/* ??? */
#define WIM_RESHDR_FLAG_FREE            0x01

/* Indicates that a file resource is a metadata resource. */
#define WIM_RESHDR_FLAG_METADATA        0x02

/* Indicates that a file resource is compressed. */
#define WIM_RESHDR_FLAG_COMPRESSED	0x04

/* ??? */
#define WIM_RESHDR_FLAG_SPANNED         0x08


/* Header at the very beginning of the WIM file. */
typedef struct WIMHeader
{ 
	/* Identifies the file as WIM file. Must be exactly
	 * {'M', 'S', 'W', 'I', 'M', 0, 0, 0}  */
	u8  magic[WIM_MAGIC_LEN]; 	

	/* size of WIM header in bytes. */
	u32 hdr_size;

	/* Version of the WIM file.  M$ provides no documentation about exactly
	 * what this field affects about the file format, other than the fact
	 * that more recent versions have a higher value. */
	u32 version;

	/* Bitwise OR of one or more of the WIM_HDR_FLAG_* defined below. */
	u32 flags;

	/* The size of the pieces that the uncompressed files were split up into
	 * when they were compressed.  This should be the same as
	 * WIM_CHUNK_SIZE.  M$ incorrectly documents this as "the size of the
	 * compressed .wim file in bytes".*/
	u32 chunk_size;
	
	/* A unique identifier for the WIM file. */
	u8  guid[WIM_GID_LEN];

	/* Part number of the WIM file in a spanned set. */
	u16 part_number;

	/* Total number of parts in a spanned set. */
	u16 total_parts;

	/* Number of images in the WIM file. */
	u32 image_count;

	/* Location, size, and flags of the lookup table. */
	WIMResourceEntry lookup_table_metadata;

	/* Location, size, and flags for the XML data. */
	WIMResourceEntry xml_metadata;

	/* Location, size, and flags for the boot metadata. */
	WIMResourceEntry boot_metametadata;

	/* The index of the bootable image in the WIM file.	If 0, there are
	 * no bootable images available. */
	u32 boot_idx; 

	/* The location of the integrity table used to verify files. */
	WIMResourceEntry integrity;

	/* Reserved for future disuse */
	u8 unused[WIM_UNUSED_LEN];
} WIMHeader;

/* Flags for the `flags' field of the WIMHeader. */


/* Reserved for future use by M$ */
#define WIM_HDR_FLAG_RESERVED           0x00000001

/* Files and metadata in the WIM are compressed. */
#define WIM_HDR_FLAG_COMPRESSION        0x00000002

/* WIM is read-only. */
#define WIM_HDR_FLAG_READONLY           0x00000004

/* Resource data specified by images in this WIM may be contained in a different
 * WIM! */
#define WIM_HDR_FLAG_SPANNED            0x00000008

/* The WIM contains resources only; no filesystem metadata. */
#define WIM_HDR_FLAG_RESOURCE_ONLY      0x00000010

/* The WIM contains metadata only. */
#define WIM_HDR_FLAG_METADATA_ONLY      0x00000020

/* Lock field to prevent multiple writers from writing the WIM concurrently. */
#define WIM_HDR_FLAG_WRITE_IN_PROGRESS  0x00000040 

/* Reparse point fixup ??? */
#define WIM_HDR_FLAG_RP_FIX             0x00000080

/* Unknown compression type */
#define WIM_HDR_FLAG_COMPRESS_RESERVED  0x00010000

/* Resources within the WIM are compressed using "XPRESS" compression, which is
 * a LZ77-based compression algorithm. */
#define WIM_HDR_FLAG_COMPRESS_XPRESS    0x00020000

/* Resources within the WIM are compressed using "LZX" compression.  This is also
 * a LZ77-based algorithm. */
#define WIM_HDR_FLAG_COMPRESS_LZX       0x00040000


/* A lookup table that is used to translate the hash codes of dentries into the
 * offsets and sizes of uncompressed or compressed file resources.  It is
 * implemented as a hash table. */
typedef struct WIMLookupTable {
	struct WIMLookupTableEntry **array;
	u64 num_entries;
	u64 capacity;
} WIMLookupTable;

/* An entry in the lookup table in the WIM file. */
typedef struct WIMLookupTableEntry {

	/* The following 4 fields are read from disk. */
	WIMResourceEntry resource_entry;
	u16 part_number;
	u32 refcnt;
	u8  hash[WIM_HASH_SIZE];

	/* If file_on_disk != NULL, the file resource indicated by this lookup
	 * table entry is not in the WIM file, but rather a file on disk; this
	 * occurs for files that are added to the WIM.  In that case,
	 * file_on_disk is the name of the file in the outside filesystem. */
	char *file_on_disk;

	/* When a WIM file is written, out_refcnt starts at 0 and is incremented
	 * whenever the file resource pointed to by this lookup table entry
	 * needs to be written.  Naturally, the file resource only need to be
	 * written when out_refcnt is 0.  Incrementing it further is needed to
	 * find the correct reference count to write to the lookup table in the
	 * output file, which may be less than the regular refcnt if not all
	 * images in the WIM file are written. 
	 *
	 * output_resource_entry is the WIMResourceEntry for the position of the
	 * file resource when written to the output file. */
	u32 out_refcnt;
	WIMResourceEntry output_resource_entry;

	/* The next WIMLookupTableEntry in the hash bucket.  NULL if this is the
	 * last one. */
	struct WIMLookupTableEntry *next;
} WIMLookupTableEntry;

/* Size of each lookup table entry in the WIM file. */
#define WIM_LOOKUP_TABLE_ENTRY_DISK_SIZE 50


/* Structure for security data.  Each image in the WIM file has its own security
 * data. */
typedef struct WIMSecurityData {
	/* The total length of the security data, in bytes.  A typical size is
	 * 2048 bytes. */
	u32 total_length;

	/* The number of security descriptors in the array descriptors, below. */
	u32 num_entries;

	/* Array of sizes of the descriptors in the array descriptors. */
	u64 *sizes;

	/* Array of descriptors. */
	u8 **descriptors;
} WIMSecurityData;


/* In-memory structure for a directory entry.  Keep all the fields up until and
 * including file_name_len compatible with the on-disk structure; extra fields
 * are following those. */
typedef struct WIMDentry {
	/* Size of directory entry, in bytes.  Typical size is around 104 to 120
	 * bytes. */
	u64 length;

	/* It is possible for the length field to be 0, and in that case the
	 * on-disk directory entry will end here, meaning the actual length is
	 * 8.  This situation, which is undocumented, indicates the end of a
	 * list of sibling nodes in a directory.  */

	/* The file attributes associated with this file. */
	u32 attributes;

	/* The index of the node in the security table that contains this file's
	 * security information.  If -1, no security information exists for this
	 * file. */
	int32_t security_id;

	/* The offset, from the start of the metadata section, of this directory
	 * entry's child files.  0 if the directory entry has no children. */
	u64 subdir_offset;

	/* Reserved */
	u64 unused1;
	u64 unused2;

	/* Timestamps for the entry.  The timestamps are the number of
	 * 100-nanosecond intervals that have elapsed since 12:00 A.M., January
	 * 1st, 1601, UTC. */
	u64 creation_time;
	u64 last_access_time;
	u64 last_write_time;

	/* A hash of the file's contents. */
	u8 hash[WIM_HASH_SIZE];

	/* Identity of a reparse point. */
	u32 reparse_tag;

	/* Although M$'s documentation does not tell you this, it seems that the
	 * reparse_reserved field does not actually exist.  So the hard_link
	 * field directly follows the reparse_tag on disk. */
	//u32 reparse_reserved;

	/* If the reparse_reserved field existed, there would be a 4-byte gap
	 * here to align hard_link on an 8-byte field.  However,
	 * reparse_reserved does not actually exist, so there is no gap here. */

	/* If the file is part of a hard link set, all the directory entries in
	 * the set will share the same value for this field. */
	u64 hard_link;

	/* Number of WIMStreamEntry structures that follow this WIMDentry. */
	u16 streams;

	/* Length of short filename, in bytes. */
	u16 short_name_len;

	/* Length of file name, in bytes. */
	u16 file_name_len;

	/* Currently at offset 102.  This is the end of the on-disk structure.
	 * On disk, this is followed by the filename, not the remaining members
	 * of this structure.  After the filename, which is file_name_len bytes,
	 * there is an undocumented 2-byte gap, then the short name, which is
	 * short_name_len bytes. */

	/* Length of the filename converted into ASCII, in bytes. */
	u16 file_name_ascii_len;

	u16 short_name_ascii_len;

	/* Pointer to the short filename */
	char *short_name;

	/* Pointer to the short filename converted into ASCII. */
	char *short_name_ascii;

	/* Pointer to the filename. */
	char *file_name;

	/* Pointer to the filename converted to ASCII. */
	char *file_name_ascii;

	/* Full path to this dentry. */
	char *fullpath_ascii;
	u32   fullpath_ascii_len;

	/* Stream entries for this dentry. */
	struct WIMStreamEntry *stream_entries;

	/* Any remaining data in the dentry after the stream entries. */
	u8 *rest;

	/* Size of the remaining data pointed to by rest.*/
	u16 restsize;
	
	/* Whether this is a directory or not. */
	//bool is_directory;

	/* The parent of this directory entry. */
	struct WIMDentry *parent;

	/* Linked list of sibling directory entries. */
	struct WIMDentry *next;
	struct WIMDentry *prev;

	/* Pointer to a child of this directory entry. */
	struct WIMDentry *children;
} WIMDentry;

#define WIM_FILE_ATTRIBUTE_READONLY            0x1
#define WIM_FILE_ATTRIBUTE_HIDDEN              0x2
#define WIM_FILE_ATTRIBUTE_SYSTEM              0x4
#define WIM_FILE_ATTRIBUTE_DIRECTORY           0x10
#define WIM_FILE_ATTRIBUTE_ARCHIVE             0x20
#define WIM_FILE_ATTRIBUTE_DEVICE              0x40
#define WIM_FILE_ATTRIBUTE_NORMAL              0x80
#define WIM_FILE_ATTRIBUTE_TEMPORARY           0x100
#define WIM_FILE_ATTRIBUTE_SPARSE_FILE         0x200
#define WIM_FILE_ATTRIBUTE_REPARSE_POINT       0x400
#define WIM_FILE_ATTRIBUTE_COMPRESSED          0x800
#define WIM_FILE_ATTRIBUTE_OFFLINE             0x1000
#define WIM_FILE_ATTRIBUTE_NOT_CONTENT_INDEXED 0x2000
#define WIM_FILE_ATTRIBUTE_ENCRYPTED           0x4000
#define WIM_FILE_ATTRIBUTE_VIRTUAL             0x10000

/* Undocumented padding between the file name and the short name that are part
 * of a directory entry in the WIM file.  */
#define WIM_FILENAME_PADDING 2

/* A directory entry may be followed by a number of alternate stream entries.
 * This feature seems to be rarely used and is currently unsupported, although
 * the stream entries are read into memory. */
typedef struct WIMStreamEntry {

	/* Length of this stream entry. */
	 u64	  length;

	 /* Reserved. */
	 u64	  unused1;

	 /* Hash code to look up this entry in the lookup table. */
	 u8       hash[WIM_HASH_SIZE];

	 /* Length of the stream name. */
	 u16	  stream_name_len;

	 /* Name of the stream. */
	 char	 *stream_name;
} WIMStreamEntry;

/* Size of the WIMStreamEntry excluding the stream name. */
#define WIM_STREAM_ENTRY_DISK_SIZE (2 * sizeof(u64) + WIM_HASH_SIZE + sizeof(u16))

/* Size of the WIMDentry up to and including the file_name_len. */
#define WIM_DENTRY_DISK_SIZE (offsetof(WIMDentry, file_name_len) +  \
							sizeof(u16)) 

/* Information that is part of a metadata resource. */
typedef struct {

	/* The security data for the image. */
	WIMSecurityData      security_data;

	/* The root dentry for the image. */
	WIMDentry            root_dentry;

	/* True if the filesystem of the image has been modified.  If this is
	 * the case, the memory for the filesystem is not freed when switching
	 * to a different WIM image. */
	bool		     modified;

	bool		     written;

	/* A pointer to the lookup table entry for the metadata resource. */
	WIMLookupTableEntry *lookup_table_entry;

} WIMBootMetadata;

/* The opaque structure exposed to the wimlib API. */
typedef struct {
	/* The name of the WIM file that has been opened. */
	char                *filename;

	/* A pointer to the file indicated by @filename, opened for reading. */
	FILE                *fp;

	/* The header of the WIM file. */
	WIMHeader            hdr;

	/* The currently selected image. */
	int                  current_image;

	/* The lookup table for the WIM file. */ 
	WIMLookupTable       lookup_table;

	/* A place to keep the new lookup table that is built when we write a
	 * WIM file. */
	WIMLookupTable       new_lookup_table;

	/* Pointer to the XML data read from the WIM file. */
	u8                  *xml_data;
	
	/* Array of the boot metadata of length image_count.  Each image in the
	 * WIM has a boot metadata associated with it. */
	WIMBootMetadata     *boot_metadata;

	/* True if files names are to be printed when doing extraction. */
	bool                 print_files;

	/* The type of links to create when extracting files (hard, symbolic, or
	 * none.) */
	int                  link_type;

	/* Name of the output directory for extraction. */
	char                *output_dir;

	/* FILE pointer for the WIM file that is being written. */
	FILE                *out_fp;

	/* Pointer to the dentry that is the working directory for the WIM file */
	WIMDentry           *working_directory;
} WIMStruct;


/* Inline utility functions for WIMStructs. */
static inline WIMDentry *wim_root_dentry(WIMStruct *w)
{
	return &w->boot_metadata[w->current_image].root_dentry;
}

static inline WIMSecurityData *wim_security_data(WIMStruct *w)
{
	return &w->boot_metadata[w->current_image].security_data;
}

static inline WIMLookupTableEntry *wim_metadata_lookup_table_entry(WIMStruct *w)
{
	return w->boot_metadata[w->current_image].lookup_table_entry;
}

static inline WIMResourceEntry *wim_metadata_resource_entry(WIMStruct *w)
{
	return &w->boot_metadata[w->current_image].lookup_table_entry->resource_entry;
}

static inline WIMResourceEntry *wim_metadata_output_resource_entry(WIMStruct *w)
{
	return &w->boot_metadata[w->current_image].lookup_table_entry->output_resource_entry;
}

static inline void wim_set_current_image_unmodified(WIMStruct *w)
{
	w->boot_metadata[w->current_image].modified = false;	
	w->boot_metadata[w->current_image].written  = true;
}

static inline void wim_set_current_image_modified(WIMStruct *w)
{
	w->boot_metadata[w->current_image].modified = true;	
	w->boot_metadata[w->current_image].written  = false;
}

static inline bool wim_current_image_is_modified(const WIMStruct *w)
{
	return w->boot_metadata[w->current_image].modified;
}

static inline bool wim_current_image_exists(const WIMStruct *w)
{
	return w->current_image != WIM_NO_IMAGE;
}

/* Inline utility functions for WIMDentry's */
static inline bool dentry_is_root(const WIMDentry *dentry)
{
	return dentry->parent == dentry;
}

static inline bool dentry_is_first_sibling(const WIMDentry *dentry)
{
	return dentry_is_root(dentry) || dentry->parent->children == dentry;
}

static inline bool dentry_is_only_child(const WIMDentry *dentry)
{
	return dentry->next == dentry;
}


static inline bool dentry_is_directory(const WIMDentry *dentry)
{
	return (dentry->attributes & WIM_FILE_ATTRIBUTE_DIRECTORY) != 0;
}

static inline void dentry_set_is_directory(WIMDentry *dentry)
{
	dentry->attributes |= WIM_FILE_ATTRIBUTE_DIRECTORY;
}

static inline void dentry_set_is_regular_file(WIMDentry *dentry)
{
	dentry->attributes |= WIM_FILE_ATTRIBUTE_NORMAL;
}

static inline bool dentry_has_children(const WIMDentry *dentry)
{
	return dentry->children != NULL;
}

static inline bool dentry_is_empty_directory(const WIMDentry *dentry)
{
	return dentry_is_directory(dentry) && !dentry_has_children(dentry);
}

/* Prints a hash code field. */
static inline void print_hash(const u8 hash[])
{
	print_byte_field(hash, WIM_HASH_SIZE);
}

/* Prints a globally-unique-identifier field. */
static inline void print_guid(const u8 guid[])
{
	print_byte_field(guid, WIM_GID_LEN);
}


/* Prints the magic characters field. */
static inline void print_magic(const u8 magic[])
{
	print_string(magic, WIM_MAGIC_LEN);
}


/* wim.c */
const u8 *get_resource_entry(const u8 *p, WIMResourceEntry *entry);
u8 *put_resource_entry(u8 *p, const WIMResourceEntry *entry);
void wimlib_error(const char *format, ...);
FILE *fopen_and_get_length(const char *path, const char *mode, 
							off_t *len_ret);

/* lookup_table.c */
void lookup_table_init(WIMLookupTable *table, size_t capacity);
void lookup_table_destroy(WIMLookupTable *table);
void lookup_table_zero_out_refcnts(WIMLookupTable *table);
WIMLookupTableEntry *new_lookup_table_entry();
bool for_lookup_table_entry(WIMLookupTable *table, 
		bool (*visitor)(WIMLookupTableEntry *, void *), void *arg);
WIMLookupTableEntry *lookup_file_resource(const WIMDentry *dentry, 
						const WIMLookupTable *lookup_table);
void lookup_table_insert(WIMLookupTable *lookup_table, 
					WIMLookupTableEntry *entry);
bool print_lookup_table_entry(WIMLookupTableEntry *entry, void *ignore);
void free_lookup_table_entry(WIMLookupTableEntry *entry);
void decrement_ref_count(const WIMDentry *dentry, 
			WIMLookupTable* lookup_table);

/* read.c */
u8 *read_xml_data(FILE *fp, WIMResourceEntry *res);

bool read_boot_metadata(FILE *fp, const WIMResourceEntry *metadata,
		u32 wim_hdr_flags, WIMSecurityData *sd, WIMDentry *root_dentry);

bool read_file_resource(FILE *fp, const WIMResourceEntry *metadata, 
				u32 wim_hdr_flags, u8 file_contents[]);

/* modify.c */
bool build_dentry_tree(WIMDentry *root, const char *source_path, 
			struct stat *root_stat, WIMLookupTable* lookup_table);

/* dentry.c */
bool for_dentry_in_tree(WIMDentry *root, 
		bool (*visitor)(WIMDentry*, void*), void *args);
bool for_dentry_in_tree_depth(WIMDentry *root, 
		bool (*visitor)(WIMDentry*, void*), void *args);

 void calculate_dentry_full_path(WIMDentry *dentry);
bool do_calculate_dentry_full_path(WIMDentry *dentry, void *ignore);
void change_dentry_name(WIMDentry *dentry, const char *new_name);
bool dentry_has_name(const WIMDentry *dentry, 
			const char *name, int name_len);

void unlink_dentry(WIMDentry *dentry);
void link_dentry(WIMDentry *dentry, WIMDentry *parent);

bool print_dentry(WIMDentry *dentry, void *lookup_table);
bool print_dentry_full_path(WIMDentry *entry, void *ignore);

WIMDentry *get_dentry(WIMStruct *w, const char *path);
WIMDentry *get_parent_dentry(WIMStruct *w, const char *path);
WIMDentry *get_dentry_child_with_name(const WIMDentry *dentry, 
							const char *name);
WIMDentry *get_dentry_relative_path(WIMDentry *cur_dir, const char *path);
void init_dentry(WIMDentry *dentry, const char *name);
WIMDentry *create_dentry(const char *name);

void destroy_dentry(WIMDentry *dentry);
bool do_free_dentry(WIMDentry *dentry, void *__args);
void free_dentry(WIMDentry *dentry);
void free_dentry_tree(WIMDentry *root, WIMLookupTable *lookup_table, 
		bool verbose, bool free_root, bool decrement_refcnt);

#include "wimlib.h"



#endif

