/*
 * write.c
 *
 * Support for writing WIM files.
 *
 * wimlib - Library for working with WIM files 
 *
 * Copyright (C) 2010 Carl Thijssen
 * Copyright (C) 2012 Eric Biggers
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */
#include "wimlib_internal.h"
#include <string.h>
#include <assert.h>
#include <errno.h>

static bool zero_out_refcnts(WIMLookupTableEntry *entry, void *ignore)
{
	entry->out_refcnt = 0;
	return true;
}

INTERNAL void lookup_table_zero_out_refcnts(WIMLookupTable *table)
{
	for_lookup_table_entry(table, zero_out_refcnts, NULL);
}

/* 
 * Recursively calculates the subdir offsets for a directory tree. 
 *
 * @dentry:  The root of the directory tree.
 * @subdir_offset_p:  The current subdirectory offset; i.e., the subdirectory
 * 	offset for @dentry. 
 */
static void calculate_subdir_offsets(WIMDentry *dentry, off_t *subdir_offset_p)
{
	dentry->subdir_offset = *subdir_offset_p;
	
	WIMDentry *child = dentry->children;
	if (child != NULL) {

		/* Advance the subdir offset by the amount of space the children
		 * of this dentry take up. */
		do {
			*subdir_offset_p += child->length;
			child = child->next;
		} while (child != dentry->children);

		/* End-of-directory dentry on disk. */
		*subdir_offset_p += 8;

		/* Recursively call calculate_subdir_offsets() on all the
		 * children. */
		do {
			calculate_subdir_offsets(child, subdir_offset_p);
			child = child->next;
		} while (child != dentry->children);
	} else {
		/* On disk, childless directories have a valid subdir_offset
		 * that points to an 8-byte end-of-directory dentry.  This
		 * indicates that they are directories and not regular files.
		 * Regular files have a subdir_offset of 0. */
		if (dentry_is_directory(dentry))
			*subdir_offset_p += 8;
		else
			dentry->subdir_offset = 0;
	}
}

/* 
 * Writes the header for a WIM file.
 *
 * @hdr: 	A pointer to a WIMHeader structure that describes the header.
 * @out:	The FILE* for the output file.
 * @return:	True on success, false on failure.
 */
static bool write_header(const WIMHeader *hdr, FILE *out)
{
	u8 buf[hdr->hdr_size];
	u8 *p = buf;
	p = put_bytes(p, WIM_MAGIC_LEN, hdr->magic);
	p = put_u32(p, hdr->hdr_size);
	p = put_u32(p, hdr->version);
	p = put_u32(p, hdr->flags);
	p = put_u32(p, hdr->chunk_size);
	p = put_bytes(p, WIM_GID_LEN, hdr->guid);
	p = put_u16(p, hdr->part_number);
	p = put_u16(p, hdr->total_parts);
	p = put_u32(p, hdr->image_count);
	p = put_resource_entry(p, &hdr->lookup_table_metadata);
	p = put_resource_entry(p, &hdr->xml_metadata);
	p = put_resource_entry(p, &hdr->boot_metametadata);
	p = put_u32(p, hdr->boot_idx);
	p = put_resource_entry(p, &hdr->integrity);
	p = put_bytes(p, WIM_UNUSED_LEN, hdr->unused);
	if (fwrite(buf, 1, hdr->hdr_size, out) != hdr->hdr_size) {
		wimlib_error("Failed to write WIM header!\n");
		return false;
	}
	return true;
}

/* 
 * Writes a lookup table entry to the output file.
 *
 * @entry:   A pointer to the lookup table entry.
 * @__out:   The open output file stream to write the data to.
 * @return:  True on success, false on failure.
 */
static bool write_lookup_table_entry(WIMLookupTableEntry *entry, void *__out)
{
	/* Don't write entries that have not had file resources or metadata
	 * resources written for them. */
	if (entry->out_refcnt == 0)
		return true;
	FILE *out = (FILE*)__out;
	if (entry->output_resource_entry.flags & WIM_RESHDR_FLAG_METADATA)
		DEBUG("Writing metadata entry at %lu\n", ftello(__out));
	u8 buf[WIM_LOOKUP_TABLE_ENTRY_DISK_SIZE];
	u8 *p = buf;
	p = put_resource_entry(p, &entry->output_resource_entry);
	p = put_u16(p, entry->part_number);
	p = put_u32(p, entry->out_refcnt);
	p = put_bytes(p, sizeof(entry->hash), entry->hash);
	assert(p - buf == WIM_LOOKUP_TABLE_ENTRY_DISK_SIZE);
	if (fwrite(buf, 1, sizeof(buf), out) != sizeof(buf)) {
		wimlib_error("Failed to write lookup table entry: %s\n",
				strerror(errno));
		return false;
	}
	return true;
}

/* 
 * Writes the lookup table to the output file.
 *
 * @table:	A pointer to the lookup table.
 * @out:	The FILE* for the output file.
 * @return:	True on success, false on failure.
 */
static bool write_lookup_table(WIMLookupTable *table, FILE *out)
{
	return for_lookup_table_entry(table, write_lookup_table_entry, out);
}

/* 
 * Writes XML data to a WIM file.
 *
 * @xml_data:	A pointer to the XML data to write.
 * @xml_len:	The length of the @xml_data.
 * @out:	The FILE* for the output file.
 * @return:	True on success, false on failure.
 */
static bool write_xml_data(const u8 *xml_data, u64 xml_len, FILE *out)
{
	if (fwrite(xml_data, 1, xml_len, out) != xml_len) {
		wimlib_error("Failed to write XML data: %s\n", strerror(errno));
		return false;
	}
	return true;
}

/* 
 * Writes the security data to the output file.
 *
 * @sd:  	The security data structure.
 * @out:  	The FILE* for the output file.
 * @return:  	True on success, false on failure.
 */
static bool write_security_data(const WIMSecurityData *sd, FILE *out)
{
	DEBUG("Writing security data (total_length = %u, num_entries = %lu)\n",
			sd->total_length, sd->num_entries);
	u8 buf[sd->total_length];
	u8 *p = buf;
	p = put_u32(p, sd->total_length);
	p = put_u32(p, sd->num_entries);

	for (unsigned i = 0; i < sd->num_entries; i++)
		p = put_u64(p, sd->sizes[i]);

	for (unsigned i = 0; i < sd->num_entries; i++)
		p = put_bytes(p, sd->sizes[i], sd->descriptors[i]);

	if (fwrite(buf, 1, sd->total_length, out) != sd->total_length) {
		wimlib_error("Failed to write security data: %s\n", 
							strerror(errno));
		return false;
	}
	DEBUG("Successfully wrote security data.\n");
	return true;
}

/*
 * Copy @len bytes from @in to @out.  Returns true on success, false on failure.
 */
static bool copy_between_files(FILE *in, FILE *out, size_t len)
{
	void *buf = xmalloc(len);
	if (fread(buf, 1, len, in) != len)
		goto copy_between_files_error;

	if (fwrite(buf, 1, len, out) != len)
		goto copy_between_files_error;

	free(buf);
	return true;
copy_between_files_error:
	free(buf);
	return false;
}

/*
 * Seek the @in stream to @in_offset, then copy @len bytes from @in to @out.
 * Returns true on success, false on failure.
 */
static bool copy_between_files_with_in_offset(FILE *in, FILE *out, size_t len, 
							off_t in_offset)
{
	if (fseeko(in, in_offset, SEEK_SET) != 0)
		return false;
	return copy_between_files(in, out, len);
}

/* 
 * Writes a file resource to the output file. 
 *
 * @dentry:  The dentry for the file resource.
 * @wim_struct_p:  A pointer to the WIMStruct.  The fields of interest to this
 * 	function are the input and output file streams and the lookup table. 
 * @return true on success, false on failure. 
 */
static bool write_file_resource(WIMDentry *dentry, void *wim_struct_p)
{
	WIMStruct *w = (WIMStruct*)wim_struct_p;
	WIMLookupTable *lookup_table = &w->lookup_table;
	FILE *out = w->out_fp;
	FILE *in  = w->fp;

	/* Directories don't need file resources. */
	if (dentry_is_directory(dentry))
		return true;


	/* Get current offset in the output file. */
	off_t out_offset = ftello(out);
	if (out_offset == -1) {
		wimlib_error("Failed to get output position: %s\n",
			strerror(errno));
		return false;
	}

	/* Get the location of the file resource in the input file. */
	WIMLookupTableEntry *lookup_table_entry = lookup_file_resource(dentry, 
								lookup_table);
	bool ret;
	if (lookup_table_entry != NULL) {

		/* No need to write file resources twice.  (This indicates file
		 * resources that are part of a hard link set.) */
		if (lookup_table_entry->out_refcnt != 0)
			return true;

		/* Figure out if we can read the resource from the WIM file, or
		 * if we have to read them from disk. */
		if (lookup_table_entry->file_on_disk != NULL) {

			DEBUG("Reading file resource `%s' from disk for dentry "
					"`%s'\n", lookup_table_entry->file_on_disk,
					dentry->fullpath_ascii);
			off_t len;
			FILE *new_file = fopen_and_get_length(
					lookup_table_entry->file_on_disk, "rb", &len);
			if (new_file == NULL)
				return false;

			ret = copy_between_files(new_file, out, len);

			lookup_table_entry->output_resource_entry.flags = 0;
			lookup_table_entry->output_resource_entry.size = len;
			lookup_table_entry->output_resource_entry.original_size = len;
		} else {
			WIMResourceEntry *entry = &lookup_table_entry->resource_entry;

			ret = copy_between_files_with_in_offset(in, out, 
					entry->size, entry->offset);

			lookup_table_entry->output_resource_entry.size = 
							entry->size;

			lookup_table_entry->output_resource_entry.original_size = 
							entry->original_size;

			lookup_table_entry->output_resource_entry.flags = 
							entry->flags;
		}
		if (!ret) {
			DEBUG("Failed to copy data from `%s' to `%s': %s\n",
				lookup_table_entry->file_on_disk,
				dentry->fullpath_ascii, strerror(errno));
		}
		lookup_table_entry->out_refcnt++;
		lookup_table_entry->output_resource_entry.offset = out_offset;
	} else {
		DEBUG("Could not find lookup table entry for `%s'\n", 
				dentry->fullpath_ascii);
		ret = true;
	}
	return ret;
}

/* 
 * Writes a dentry to the output file.
 *
 * @dentry:  The dentry structure.
 * @out:     The open output file stream to write the data to.
 * @return:  True on success, false on failure.
 */
static bool write_dentry(WIMDentry *dentry, FILE *out)
{
	u8 buf[dentry->length];
	u8 *p = buf;
	assert(dentry->length >= WIM_DENTRY_DISK_SIZE);
	p = put_u64(p, dentry->length);
	p = put_u32(p, dentry->attributes);
	p = put_u32(p, dentry->security_id);
	p = put_u64(p, dentry->subdir_offset);
	p = put_u64(p, dentry->unused1);
	p = put_u64(p, dentry->unused2);
	p = put_u64(p, dentry->creation_time);
	p = put_u64(p, dentry->last_access_time);
	p = put_u64(p, dentry->last_write_time);
	p = put_bytes(p, WIM_HASH_SIZE, dentry->hash);
	p = put_u32(p, dentry->reparse_tag);
	p = put_u64(p, dentry->hard_link);
	p = put_u16(p, dentry->streams);
	p = put_u16(p, dentry->short_name_len);
	p = put_u16(p, dentry->file_name_len);
	p = put_bytes(p, dentry->file_name_len, (u8*)dentry->file_name);
	p = put_u16(p, 0); /* filename padding, 2 bytes. */
	p = put_bytes(p, dentry->short_name_len, (u8*)dentry->short_name);
	WIMStreamEntry *sentry = dentry->stream_entries;
	unsigned num_streams = dentry->streams;
	while (num_streams--) {
		p = put_u64(p, sentry->length);
		p = put_u64(p, sentry->unused1);
		p = put_bytes(p, WIM_HASH_SIZE, sentry->hash);
		p = put_u16(p, sentry->stream_name_len);
		p = put_bytes(p, sentry->stream_name_len, (u8*)sentry->stream_name);
		sentry++;
	}
	if (fwrite(buf, 1, (p - buf), out) != (p - buf))
		goto write_dentry_error;


	if (dentry->restsize != 0) {
		/* Write the remaining bytes pointed to by dentry->rest.  If
		 * it's NULL, we added this dentry ourselves and are just going
		 * to write zeroes (since presumably it's just padding anyway).
		 * */
		if (dentry->rest == NULL) {
			memset(p, 0, dentry->restsize);
			if (fwrite(p, 1, dentry->restsize, out) != dentry->restsize)
				goto write_dentry_error;
		} else {
			if (fwrite(dentry->rest, 1, dentry->restsize, out) != 
					dentry->restsize)
				goto write_dentry_error;
		}

	}
	assert(p - buf + dentry->restsize == dentry->length);
	return true;
write_dentry_error:
	wimlib_error("Failed to write dentry: %s\n", strerror(errno));
	return false;
}

/* Writes an end-of-directory entry to @out. */
static bool write_end_of_directory_entry(FILE *out)
{
	u8 buf[8];
	put_u64(buf, 0);
	if (fwrite(buf, 1, sizeof(buf), out) != sizeof(buf)) {
		wimlib_error("Failed to write end-of-directory entry: %s\n", 
				strerror(errno));
		return false;
	}
	return true;
}


/* Recursive function that writes a dentry tree rooted at @tree, not including
 * @tree itself, which has already been written, except in the case of the root
 * dentry, which is written right away, along with an end-of-directory entry. */
static bool write_dentry_tree(WIMDentry *tree, FILE *out)
{
	if (dentry_is_root(tree)) {
		if (!write_dentry(tree, out))
			return false;
		if (!write_end_of_directory_entry(out))
			return false;
	}

	/* Nothing to do for a regular file. */
	if (!dentry_is_directory(tree))
		return true;

	/* Write child dentries and end-of-directory entry. */
	WIMDentry *child = tree->children;
	if (child != NULL) {
		do {
			if (!write_dentry(child, out))
				return false;
			child = child->next;
		} while (child != tree->children);
	}

	if (!write_end_of_directory_entry(out))
		return false;

	/* Recurse on children. */
	if (child != NULL) {
		do {
			if (!write_dentry_tree(child, out))
				return false;
			child = child->next;
		} while (child != tree->children);
	}
	return true;
}

/* 
 * Writes the WIM to a file.
 *
 * @w:  A WIMStruct that specifies a WIM file that has been opened for reading.
 * 	There may be in-memory changes made to it, which are then reflected in 
 * 	the output file.
 * @out_wim_path:  The file to write the WIM to.
 * @return:  True on success, false on failure.
 */
bool wimlib_write(WIMStruct *w, const char *out_wim_path, 
				bool current_image_only)
{
	if (w->current_image == WIM_NO_IMAGE && !wimlib_select_image(w, 1))
		return false;

	u32 original_image = w->current_image;

	FILE *out = fopen(out_wim_path, "wb");
	if (out == NULL) {
		wimlib_error("Failed to open the file `%s' for writing!\n", 
								out_wim_path);
		return false;
	}

	w->out_fp = out;

	/* Create and write a preliminary header based on the current header. It
	 * will be overwritten later. */
	WIMHeader hdr;
	memcpy(&hdr, &w->hdr, sizeof(WIMHeader));
	if (!write_header(&hdr, out))
		return false;

	/* Set all lookup table entries to not written. This flag indicates that
	 * their file resources have not yet been written to the WIM file. */
	lookup_table_zero_out_refcnts(&w->lookup_table);

	/* Write the file resources for each image, or only for the current
	 * image if @current_image_only was passed as a parameter. */

	DEBUG("current_image_only = %u, w->current_image = %u, original_image = %u\n",
			current_image_only, w->current_image, original_image);
	for (u32 i = 0; i < w->hdr.image_count; i++) {
		if (current_image_only) {
			if (i != w->current_image)
				continue;
		} else {
			if (!wimlib_select_image(w, i + 1))
				return false;
		}
		w->boot_metadata[i].written = true;
		DEBUG("Writing file resources for image %u\n", i + 1);
		if (!for_dentry_in_tree(wim_root_dentry(w), write_file_resource, w)) 
		{
			wimlib_error("Failed to write file resources "
					"to `%s'!\n", out_wim_path);
			return false;
		}
	}

	off_t boot_metadata_offset = ftello(out);
	off_t cur_offset = boot_metadata_offset;

	/* Write the boot metadata for each image, or only for the current image
	 * if @current_image_only was passed as a paramater. */
	for (u32 i = 0; i < w->hdr.image_count; i++) {

		if (current_image_only) {
			if (i != w->current_image)
				continue;
		} else {
			if (!wimlib_select_image(w, i + 1))
				return false;
		}

		DEBUG("Writing the boot metadata for image %u.\n", i + 1);

		if (!write_security_data(wim_security_data(w), out))
			return false;

		/* Write the dentries. */
		off_t offset = wim_security_data(w)->total_length + 
					wim_root_dentry(w)->length + 8;
		DEBUG("Calculating subdir offsets for image %u.\n", i + 1);
		calculate_subdir_offsets(wim_root_dentry(w), &offset);

		DEBUG("Writing dentry tree for image %u.\n", i + 1);
		if (!write_dentry_tree(wim_root_dentry(w), out))
			return false;

		DEBUG("Updating resource entry for image %u's boot metadata.\n",
				i + 1);
		/* Update the resource entry for this image's boot
		 * metadata. */
		WIMResourceEntry *entry = wim_metadata_output_resource_entry(w);
		wim_metadata_lookup_table_entry(w)->out_refcnt++;
		entry->offset        = cur_offset;
		cur_offset           = ftello(out);
		entry->size          = cur_offset - entry->offset;
		entry->original_size = entry->size;
		entry->flags         = WIM_RESHDR_FLAG_METADATA;
	}

	off_t lookup_table_offset = ftello(out);


	DEBUG("Writing lookup table.\n");
	/* Write the lookup table. */
	if (!write_lookup_table(&w->lookup_table, out))
		return false;

	off_t xml_data_offset = ftello(out);

	/* Write the XML data. */
	if (w->xml_data == NULL)
		if ((w->xml_data = read_xml_data(w->fp, &w->hdr.xml_metadata)) == NULL)
			return false;

	if (!write_xml_data(w->xml_data, w->hdr.xml_metadata.size, out))
		return false;

	/* Calculate metadata information, update the header, and write the new
	 * header at the start of the file.  */
	off_t lookup_table_size = xml_data_offset - lookup_table_offset;
	hdr.xml_metadata.offset                 = xml_data_offset;
	hdr.lookup_table_metadata.offset        = lookup_table_offset;
	hdr.lookup_table_metadata.size          = lookup_table_size;
	hdr.lookup_table_metadata.original_size = lookup_table_size;
	hdr.lookup_table_metadata.flags         &= ~WIM_RESHDR_FLAG_COMPRESSED;

	/* 
	 * In the WIM header, here is room for one resource entry for boot
	 * metadata.  This entry was zeroed out on the 5-image install.wim from
	 * the Windows 7 DVD, and you have to search the lookup table to find
	 * the resource entries for the metadata resources.  However, Windows PE
	 * was unable to read a 2-image WIM with this entry zeroed out.  I'm not
	 * sure exactly what it's used for.  For now, just put the resource
	 * entry for the first metadata resource there.
	 */
	unsigned hdr_boot_metametadata_image_idx;
	if (current_image_only)
		hdr_boot_metametadata_image_idx = w->current_image;
	else
		hdr_boot_metametadata_image_idx = 0;
	memcpy(&hdr.boot_metametadata, wim_metadata_output_resource_entry(w),
			sizeof(WIMResourceEntry));

	/* Set image count correctly. */
	if (current_image_only) {
		if (hdr.boot_idx != 0)
			hdr.boot_idx = 1;
		hdr.image_count = 1;
	}

	fseeko(out, 0, SEEK_SET);
	if (!write_header(&hdr, out))
		return false;

	if (fclose(out) != 0) {
		wimlib_error("Failed to close the file `%s': %s\n",
				out_wim_path, strerror(errno));
		return false;
	}

	if (!wimlib_select_image(w, original_image + 1)) {
		wimlib_error("Failed to switch back to image %u!\n",
				original_image + 1);
		return false;
	}

	/* Done writing WIM file! */
	return true;
}
